from itertools import izip 

from .. import config, names,  prims, syntax

from ..builder import mk_prim_fn 
from ..ndtypes import (Type, 
                       array_type, closure_type, tuple_type, type_conv, 
                       Bool, IntT, Int64,  ScalarT, ArrayT,  
                       NoneType, NoneT, Unknown, UnknownT, 
                       TypeValueT, 
                       TupleT, make_tuple_type, 
                       lower_rank, make_array_type, 
                       ClosureT)

from ..syntax import adverb_helpers
from ..syntax import (UntypedFn, TypedFn, Closure,  Var, Map, Expr, 
                      ActualArgs, FormalArgs, MissingArgsError, TooManyArgsError)

from ..syntax.helpers import (get_type, get_types,  get_elt_types, 
                              one_i64, zero_i64, none, true, false, 
                              gen_data_arg_names, unwrap_constant)
from ..syntax.wrappers import build_untyped_prim_fn, build_untyped_cast_fn


from helpers import untyped_identity_function, make_typed_closure, _get_closure_type, _get_fundef
from linearize_args import linearize_actual_args, flatten_actual_args
from local_inference import LocalTypeInference 
from var_map import VarMap 


class TypeInferenceError(Exception):
  def __init__(self, msg, expr):
    self.msg = msg
    self.expr = expr 

_invoke_type_cache = {}
def invoke_result_type(fn, arg_types):
  if fn.__class__ is TypedFn:
    assert isinstance(arg_types, (list, tuple))
    assert len(arg_types) == len(fn.input_types), \
        "Type mismatch between expected inputs %s and %s" % \
        (fn.input_types, arg_types)
    assert all(t1 == t2 for (t1,t2) in zip(arg_types, fn.input_types))
    return fn.return_type

  if isinstance(arg_types, (list, tuple)):
    arg_types = ActualArgs(arg_types)
  key = (fn, arg_types)
  if key in _invoke_type_cache:
    return _invoke_type_cache[key]

  if fn.__class__ is ClosureT:
    closure_set = closure_type.ClosureSet(fn)
  else:
    assert isinstance(fn, closure_type.ClosureSet), \
        "Invoke expected closure, but got %s" % (fn,)
    closure_set = fn

  result_type = Unknown
  for closure_t in closure_set.closures:
    typed_fundef = specialize(closure_t, arg_types)
    result_type = result_type.combine(typed_fundef.return_type)
  _invoke_type_cache[key] = result_type
  return result_type

class TypeInference(LocalTypeInference):
  
  
  def transform_args(self, args, flat = False):
    if isinstance(args, (list, tuple)):
      return self.transform_expr_list(args)
    else:
      new_args = args.transform(self.transform_expr)
      if flat:
        return flatten_actual_args(new_args)
      else:
        return new_args

  def transform_keywords(self, kwds_dict):
    if kwds_dict is None:
      return {}
    else:
      result = {}
      for (k,v) in kwds_dict.iteritems():
        result[k] = self.transform_expr(v)
      return result

  
  def transform_fn(self, f):
  
    """
    If you're calling a Type, turn it into a wrapper function around the Cast expression. 
    If you're calling a Prim, turn it into a wrapper function around the PrimCall expression.
    """
    if isinstance(f, Type):
      expr = build_untyped_cast_fn(f)
    elif isinstance(f, prims.Prim):
      expr = build_untyped_prim_fn(f)
    else:
      expr = f 
    return self.transform_expr(expr)
  
  def transform_Call(self, expr):
    closure = self.transform_fn(expr.fn)
    args = self.transform_args(expr.args)
    if closure.type.__class__ is TypeValueT:
      assert args.__class__ is  ActualArgs
      assert len(args.positional) == 1
      assert len(args.keywords) == 0
      assert args.starargs is None 
      return self.cast(args.positional[0], closure.type.type)
    
    untyped_fn, args, arg_types = linearize_actual_args(closure, args)
    typed_fn = specialize(untyped_fn, arg_types)
    return syntax.Call(typed_fn, tuple(args), typed_fn.return_type)

  def transform_PrimCall(self, expr):
    args = self.transform_args(expr.args)

    arg_types = get_types(args)

    if all(isinstance(t, ScalarT) for t in arg_types):
      upcast_types = expr.prim.expected_input_types(arg_types)
      result_type = expr.prim.result_type(upcast_types)
      return syntax.PrimCall(expr.prim, args, type = result_type)
    elif expr.prim == prims.is_:
      if arg_types[0] != arg_types[1]:
        return false
      elif arg_types[0] == NoneType:
        return true
      else:
        return syntax.PrimCall(prims.is_, args, type = Bool)
    elif all(t.rank == 0 for t in arg_types):
      # arguments should then be tuples
      assert len(arg_types) == 2, \
        "Expected two arguments but got [%s] in %s" % (", ".join(str(t) for t in arg_types), expr)
      xt, yt = arg_types
      x, y = args
      assert xt.__class__ is TupleT, \
        "Unexpected argument types (%s,%s) for operator %s" % (xt, yt, expr.prim)
      assert yt.__class__ is TupleT, \
        "Unexepcted argument types (%s,%s) for operator %s" % (xt, yt, expr.prim)
      x_elts = self.tuple_elts(x)
      y_elts = self.tuple_elts(y)
      nx = len(x_elts)
      ny = len(y_elts)
      if expr.prim is prims.add:
        return self.tuple( tuple(x_elts) + tuple(y_elts) )
      elif expr.prim is prims.equal:
        assert len(x_elts) == len(y_elts), \
          "Can't compare tuple of unequal lengths %d and %d" % (nx, ny)
        result = true  
        for (xi, yi) in zip(x_elts, y_elts):
          elts_eq = syntax.PrimCall(prims.equal, (xi, yi), type=Bool)
          result = syntax.PrimCall(prims.logical_and, (result, elts_eq), type=Bool) 
        return result  
      elif expr.prim is prims.not_equal:
        assert len(x_elts) == len(y_elts), \
          "Can't compare tuple of unequal lengths %d and %d" % (nx, ny)
        result = false  
        for (xi, yi) in zip(x_elts, y_elts):
          elts_eq = syntax.PrimCall(prims.not_equal, (xi, yi), type=Bool)
          result = syntax.PrimCall(prims.logical_or, (result, elts_eq), type=Bool) 
        return result  
      
      else:
        assert False, "Unsupported tuple operation %s" % expr  
    else:
      assert all(t.__class__ is not NoneT for t in arg_types), \
        "Invalid argument types for prim %s: %s" % (expr.prim, arg_types,)
      elt_types = [t.elt_type if isinstance(t, ArrayT) else t for t in arg_types]
      typed_prim_fn = mk_prim_fn(expr.prim, elt_types)
      max_rank = adverb_helpers.max_rank(arg_types)
      if max_rank == 0:
        return self.call(typed_prim_fn, args)
      elif max_rank == 1:
        axis = zero_i64 
      else:
        axis = none 
      result_t = make_array_type(typed_prim_fn.return_type, max_rank)
      return Map(fn = typed_prim_fn, args = args, type = result_t, axis = axis)
      

  def transform_Zip(self, expr):
    assert isinstance(expr.values, (list,tuple)), "Expected multiple values but got %s" % expr.values
    typed_values = self.transform_expr_tuple(expr.values)
    # if you're just zipping tuples together, 
    # then the result can also be a tuple
    if all(isinstance(v.type, TupleT) for v in typed_values):
      # keep the shortest tuple
      n = min(len(v.type.elt_types) for v in typed_values)
      zip_inputs = []
      for v in typed_values:
        zip_inputs.append(self.tuple_elts(v)[:n])
      return self.tuple([self.tuple(group) for group in zip(*zip_inputs)])
    
    # if any are tuples, that puts a max length on any sequence
    elif any(isinstance(v.type, TupleT) for v in typed_values):
      # keep the shortest tuple
      n = min(len(v.type.elt_types) for v in typed_values if isinstance(v.type, TupleT))
      assert False, "Zipping of mixed tuples and arrays not yet supported"  
    else:
      assert all(isinstance(v.type, ArrayT) for v in typed_values), \
        "Expected all inputs to zip to be arrays but got %s" % \
        ", ".join(str(v.type) for v in typed_values)
      
      elt_t = make_tuple_type([v.type.elt_type for v in typed_values])  
      result_t = make_array_type(elt_t, 1)
      def tupler(*args):
        return args 
      from ..frontend import ast_conversion
      untyped = ast_conversion.translate_function_value(tupler)
      typed = specialize(untyped, [v.type.elt_type for v in typed_values])
      result = Map(fn = typed, args = typed_values, axis = zero_i64, type = result_t)
      assert False, "Materializing the result of zipping arrays not yet supported"
      #return result
 
  def transform_IndexMap(self, expr):
    shape = self.transform_expr(expr.shape)
    if not isinstance(shape.type, TupleT):
      assert isinstance(shape.type, ScalarT), "Invalid shape for IndexMap: %s : %s" % (shape, shape.type)
      shape = self.tuple((shape,))
    closure = self.transform_fn(expr.fn)
    shape_t = shape.type
    if isinstance(shape_t, IntT):
      shape = self.cast(shape, Int64)
      n_indices = 1
    else:
      assert isinstance(shape_t, TupleT), "Expected shape to be tuple, instead got %s" % (shape_t,)
      assert all(isinstance(t, ScalarT) for t in shape_t.elt_types)
      n_indices = len(shape_t.elt_types)
      if not all(t == Int64 for t in shape_t.elt_types):
        elts = tuple(self.cast(elt, Int64) for elt in self.tuple_elts(shape))
        shape = self.tuple(elts)
    result_type, typed_fn = specialize_IndexMap(closure.type, n_indices)
    return syntax.IndexMap(shape = shape, 
                           fn = make_typed_closure(closure, typed_fn), 
                           type = result_type)
  
  def transform_IndexReduce(self, expr):
    shape = self.transform_expr(expr.shape)
    map_fn_closure = self.transform_fn(expr.fn if expr.fn else untyped_identity_function)
    combine_closure = self.transform_fn(expr.combine)
    init = self.transform_if_expr(expr.init)
    shape_t = shape.type
    if isinstance(shape_t, IntT):
      shape = self.cast(shape, Int64)
      n_indices = 1
    else:
      assert isinstance(shape_t, TupleT)
      assert all(isinstance(t, ScalarT) for t in shape_t.elt_types)
      n_indices = len(shape_t.elt_types)
      if not all(t == Int64 for t in shape_t.elt_types):
        elts = tuple(self.cast(elt, Int64) for elt in self.tuple_elts(shape))
        shape = self.tuple(elts)
    result_type, typed_fn, typed_combine = \
      specialize_IndexReduce(map_fn_closure.type, combine_closure, n_indices, init)
    if not self.is_none(init):
      init = self.cast(init, result_type)
    return syntax.IndexReduce(shape = shape, 
                              fn = make_typed_closure(map_fn_closure, typed_fn),
                              combine = make_typed_closure(combine_closure, typed_combine),
                              init = init,  
                              type = result_type)
  
  def transform_Map(self, expr):
    closure = self.transform_fn(expr.fn)
    new_args = self.transform_args(expr.args, flat = True)
    arg_types = get_types(new_args)
    assert len(arg_types) > 0, "Map requires array arguments"
    # if all arguments are scalars just handle map as a regular function call
    if all(isinstance(t, ScalarT) for t in arg_types):
      return self.invoke(closure, new_args)
    # if any arguments are tuples then all of them should be tuples of same len
    elif any(isinstance(t, TupleT) for t in arg_types):
      assert all(isinstance(t, TupleT) for t in arg_types), \
        "Map doesn't support input types %s" % (arg_types,)
      nelts = len(arg_types[0].elt_types)
      assert all(len(t.elt_types) == nelts for t in arg_types[1:]), \
       "Tuple arguments to Map must be of same length"
      zipped_elts = []
      for i in xrange(nelts):
        zipped_elts.append([self.tuple_proj(arg,i) for arg in new_args])
      return self.tuple([self.invoke(closure, elts) for elts in zipped_elts])
    axis = self.transform_if_expr(expr.axis)
    axes = self.normalize_axes(new_args, axis)
    result_type, typed_fn = specialize_Map(closure.type, arg_types, axes)
    return syntax.Map(fn = make_typed_closure(closure, typed_fn),
                       args = new_args,
                       axis = axis,
                       type = result_type)


  
    
  def transform_Reduce(self, expr):
    assert len(expr.args) > 0, "Can't have Reduce without any arguments %s" % expr 
    new_args = self.transform_args(expr.args, flat = True)
    arg_types = get_types(new_args)
    axis = self.transform_if_expr(expr.axis)

    map_fn = self.transform_fn(expr.fn if expr.fn else untyped_identity_function) 
    
    
    # if there aren't any arrays, just treat this as a function call
    if all(isinstance(t, ScalarT) for t in arg_types):
      return self.invoke(map_fn, new_args)
    
    combine_fn = self.transform_fn(expr.combine)
    if self.is_none(expr.init):
      init = none 
    else: 
      init = self.transform_expr(expr.init)
      
    axes = self.normalize_axes(new_args, axis)
    result_type, typed_map_fn, typed_combine_fn = \
      specialize_Reduce(map_fn.type,
                        combine_fn.type,
                        arg_types, 
                        axes, 
                        init.type)  

    typed_map_closure = make_typed_closure (map_fn, typed_map_fn)
    typed_combine_closure = make_typed_closure(combine_fn, typed_combine_fn)
    
    # if we encounter init = 0 for a Reduce which produces an array
    # then need to broadcast to get an initial value of the appropriate rank 
    if init.type.__class__ is not NoneT and \
       init.type != result_type and \
       array_type.rank(init.type) < array_type.rank(result_type):
      assert len(new_args) == 1, "Can't have more than one arg in " % expr  
      arg = new_args[0]
      first_elt = self.slice_along_axis(arg, axis, zero_i64)
      first_combine = specialize(combine_fn, (init.type, first_elt.type))
      
      first_combine_closure = make_typed_closure(combine_fn, first_combine)
      init = self.call(first_combine_closure, (init, first_elt))
      slice_rest = syntax.Slice(start = one_i64, stop = none, step = one_i64, 
                                   type = array_type.SliceT(Int64, NoneType, Int64))
      rest = self.slice_along_axis(arg, axis, slice_rest)
      new_args = (rest,)  
    
    return syntax.Reduce(fn = typed_map_closure,
                         combine = typed_combine_closure,
                         args = new_args,
                         axis = axis,
                         type = result_type,
                         init = init)

  def transform_Scan(self, expr):
    map_fn = self.transform_fn(expr.fn if expr.fn else untyped_identity_function)
    combine_fn = self.transform_fn(expr.combine)
    emit_fn = self.transform_fn(expr.emit)
    new_args = self.transform_args(expr.args, flat = True)
    arg_types = get_types(new_args)
    
    init = self.transform_expr(expr.init) if expr.init else None
    
    init_type = get_type(init) if init else None
    
    axis = self.transform_if_expr(expr.axis)
    axes = self.normalize_axes(new_args, axis)
    result_type, typed_map_fn, typed_combine_fn, typed_emit_fn = \
        specialize_Scan(map_fn.type, combine_fn.type, emit_fn.type,
                        arg_types, axes, init_type)
    map_fn.fn = typed_map_fn
    combine_fn.fn = typed_combine_fn
    emit_fn.fn = typed_emit_fn
    return syntax.Scan(fn = make_typed_closure(map_fn, typed_map_fn),
                       combine = make_typed_closure(combine_fn,
                                                    typed_combine_fn),
                       emit = make_typed_closure(emit_fn, typed_emit_fn),
                       args = new_args,
                       axis = axis,
                       type = result_type,
                       init = init)

  def transform_OuterMap(self, expr):
    closure = self.transform_fn(expr.fn)
    new_args = self.transform_args (expr.args, flat = True)
    arg_types = get_types(new_args)
    n_args = len(arg_types)
    assert n_args > 0
    axis = self.transform_if_expr(expr.axis)
    axes = self.normalize_axes(new_args, axis)
    result_type, typed_fn = specialize_OuterMap(closure.type, arg_types, axes)
    result = syntax.OuterMap(fn = make_typed_closure(closure, typed_fn),
                             args = new_args,
                             axis = axis,
                             type = result_type)
    return result 

  
  


def infer_types(untyped_fn, types):
  """
  Given an untyped function and input types, propagate the types through the
  body, annotating the AST with type annotations.

  NOTE: The AST won't be in a correct state until a rewrite pass back-propagates
  inferred types throughout the program and inserts adverbs for scalar operators
  applied to arrays
  """
  
  var_map = VarMap()
  typed_args = untyped_fn.args.transform(rename_fn = var_map.rename)
  if untyped_fn.args.starargs:
    assert typed_args.starargs, "Missing star-args in call to %s(%s)" % (untyped_fn.name, typed_args,)

  unbound_keywords = []
  def keyword_fn(local_name, value):
    unbound_keywords.append(local_name)

    if isinstance(value, Expr):
      value = unwrap_constant(value) 
    return type_conv.typeof(value)

  try: 
    tenv = typed_args.bind(types,
                           keyword_fn = keyword_fn,
                           starargs_fn = tuple_type.make_tuple_type)
  except (MissingArgsError, TooManyArgsError) as e:
    e.fn_name = untyped_fn.name 
    raise e
  except: 
    print "Error while calling %s with types %s" % (untyped_fn, types)
    raise
  # keep track of the return
  tenv['$return'] = Unknown
  annotator = TypeInference(tenv, var_map)
  body = annotator.transform_block(untyped_fn.body)
  arg_names = [local_name for local_name
               in
               typed_args.nonlocals + tuple(typed_args.positional)
               if local_name not in unbound_keywords]
  if len(unbound_keywords) > 0:
    default_assignments = []
    from ..frontend.ast_conversion import value_to_syntax 
    for local_name in unbound_keywords:
      t = tenv[local_name]
      python_value = typed_args.defaults[local_name]
      var = Var(local_name, type = t)
      parakeet_value = value_to_syntax(python_value)
      parakeet_value.type = t 
      stmt = syntax.Assign(var, parakeet_value)
      default_assignments.append(stmt)
    body = default_assignments + body

  input_types = tuple([tenv[arg_name] for arg_name in arg_names])

  # starargs are all passed individually and then packaged up
  # into a tuple on the first line of the function
  if typed_args.starargs:
    local_starargs_name = typed_args.starargs

    starargs_t = tenv[local_starargs_name]
    assert starargs_t.__class__ is TupleT, \
        "Unexpected starargs type %s" % starargs_t
    extra_arg_vars = []
    for (i, elt_t) in enumerate(starargs_t.elt_types):
      arg_name = "%s_elt%d" % (names.original(local_starargs_name), i)
      tenv[arg_name] = elt_t
      arg_var = Var(name = arg_name, type = elt_t)
      arg_names.append(arg_name)
      extra_arg_vars.append(arg_var)
    input_types = input_types + starargs_t.elt_types
    tuple_lhs = Var(name = local_starargs_name, type = starargs_t)
    tuple_rhs = syntax.Tuple(elts = extra_arg_vars, type = starargs_t)
    stmt = syntax.Assign(tuple_lhs, tuple_rhs)
    body = [stmt] + body

  return_type = tenv["$return"]
  # if nothing ever gets returned, then set the return type to None
  if return_type.__class__ is  UnknownT:
    body.append(syntax.Return(none))
    tenv["$return"] = NoneType
    return_type = NoneType

  return TypedFn(
    name = names.refresh(untyped_fn.name),
    body = body,
    arg_names = arg_names,
    input_types = input_types,
    return_type = return_type,
    type_env = tenv)

def _specialize(fn, arg_types, return_type = None):
  """
  Do the actual work of type specialization, whereas the wrapper 'specialize'
  pulls out untyped functions from closures, wraps argument lists in ActualArgs
  objects and performs memoization
  """
  if fn.__class__ is TypedFn:
    return fn
  typed_fundef = infer_types(fn, arg_types)
  from rewrite_typed import rewrite_typed
  return rewrite_typed(typed_fundef, return_type)

def specialize(fn, arg_types, return_type = None):
  if fn.__class__ is TypedFn:
    assert len(fn.input_types) == len(arg_types)
    assert all(t1 == t2 for t1,t2 in izip(fn.input_types, arg_types))
    if return_type is not None: assert fn.return_type == return_type 
    return fn
  elif isinstance(arg_types, (list, tuple)):
    arg_types = ActualArgs(arg_types)
    
  closure_t = _get_closure_type(fn)
  key = arg_types, return_type
  
  if key in closure_t.specializations:
    return closure_t.specializations[key]

  if config.print_before_specialization:
    if return_type:
      print "==== Specializing", fn, "for input types", arg_types, "and return type", return_type
    else:  
      print "=== Specializing", fn, "for types", arg_types 
  
  full_arg_types = arg_types.prepend_positional(closure_t.arg_types)
  fundef = _get_fundef(closure_t.fn)
  typed =  _specialize(fundef, full_arg_types, return_type)
  closure_t.specializations[key] = typed

  if config.print_specialized_function:
    if return_type:
      print "=== Specialized %s for input types %s and return type %s ==="  % \
          (fundef.name, full_arg_types, return_type)
    else:
      print "=== Specialized %s for input types %s ==="  % \
          (fundef.name, full_arg_types)
    
    print
    print repr(typed)
    print

  return typed

def infer_return_type(untyped, arg_types):
  """
  Given a function definition and some input types, gives back the return type
  and implicitly generates a specialized version of the function.
  """
  return specialize(untyped, arg_types).return_type

def specialize_IndexMap(fn, n_indices):
  idx_type = make_tuple_type( (Int64,) * n_indices) if n_indices > 1 else Int64
  typed_fn = specialize(fn, (idx_type,))
  result_type = array_type.increase_rank(typed_fn.return_type, n_indices)
  return result_type, typed_fn

def specialize_IndexReduce(fn, combine, n_indices, init = None):
  idx_type = make_tuple_type( (Int64,) * n_indices) if n_indices > 1 else Int64
  if init is None or init.type.__class__ is  NoneT:
    typed_fn = specialize(fn, (idx_type,))
  else:
    typed_fn = specialize(fn, (idx_type,), return_type = init.type)
  elt_type = typed_fn.return_type
  typed_combine = specialize(combine, (elt_type, elt_type))
  return elt_type, typed_fn, typed_combine 

def peel_adverb_input_types(array_types, axes):
  assert len(array_types) == len(axes), \
    "Mismatch between types %s and axes %s" % (array_types, axes)
  result = []
  for t, axis in zip(array_types, axes):
    if axis is None:
      result.append(array_type.elt_type(t))
    else:
      result.append(array_type.lower_rank(t,1))
  return tuple(result)

def rank(t):
  if t.__class__ is ArrayT:
    return t.rank 
  else:
    return 0 

def increase_adverb_output_rank(array_types, axes, elt_result_type, cartesian_product = False):
  delta_rank = 0
  assert len(array_types) == len(axes), \
    "Mismatch between types %s and axes %s" % (array_types, axes)
  for (t,axis) in zip(array_types, axes):
    if axis is None: r = rank(t)
    else: r = 1
    if cartesian_product: delta_rank += r 
    else: delta_rank = max(r, delta_rank)
  return array_type.increase_rank(elt_result_type, delta_rank)

def specialize_Map(map_fn, array_types, axes):
  
  elt_types = peel_adverb_input_types(array_types, axes)
  typed_map_fn = specialize(map_fn, elt_types)
  elt_result_t = typed_map_fn.return_type
  result_t = increase_adverb_output_rank(array_types, axes, elt_result_t)
  return result_t, typed_map_fn

def specialize_Reduce(map_fn, combine_fn, array_types, axes, init_type = None):
  _, typed_map_fn = specialize_Map(map_fn, array_types, axes)
  elt_type = typed_map_fn.return_type
  if init_type is None or init_type.__class__ is NoneT:
    acc_type = elt_type
  else:
    acc_type = init_type
  typed_combine_fn = specialize(combine_fn, [acc_type, elt_type])
  new_acc_type = typed_combine_fn.return_type
  if new_acc_type != acc_type:
    typed_combine_fn = specialize(combine_fn, [new_acc_type, elt_type])
    new_acc_type = typed_combine_fn.return_type
  return new_acc_type, typed_map_fn, typed_combine_fn

def specialize_Scan(map_fn, combine_fn, emit_fn, array_types, axes,  init_type = None):
  acc_type, typed_map_fn, typed_combine_fn = \
      specialize_Reduce(map_fn, combine_fn, array_types, axes, init_type)
  typed_emit_fn = specialize(emit_fn, [acc_type])
  elt_result_t = typed_emit_fn.return_type
  result_t = increase_adverb_output_rank(array_types, axes, elt_result_t)
  return result_t, typed_map_fn, typed_combine_fn, typed_emit_fn

def specialize_OuterMap(fn, array_types, axes):
  elt_types = peel_adverb_input_types(array_types, axes)
  typed_map_fn = specialize(fn, elt_types)
  elt_result_t = typed_map_fn.return_type
  result_t = increase_adverb_output_rank(array_types, axes, elt_result_t, cartesian_product = True)
  return result_t, typed_map_fn




from dsltools import NestedBlocks

from .. import names, syntax
from ..ndtypes import NoneType 
from ..syntax import TypedFn
from ..builder import Builder 


def fresh_builder(fn):
  blocks = NestedBlocks()
  blocks.push(fn.body)
  return Builder(type_env = fn.type_env, blocks = blocks)
  
def build_fn(input_types, 
             return_type = NoneType, 
             name = None, 
             input_names = None):
  n_inputs = len(input_types)
  if input_names is None:
    input_names = syntax.gen_data_arg_names(n_inputs)
  assert len(input_names) == n_inputs 
  if name is None:
    name = 'f'
  name = names.refresh(name)
 
  f = TypedFn(name = name, 
              arg_names = input_names, 
              body = [], 
              type_env = dict(zip(input_names, input_types)),
              input_types = input_types, 
              return_type = return_type) 
              
  builder = fresh_builder(f)
  input_vars = builder.input_vars(f) 
  return f, builder, input_vars  


_identity_fn_cache = {}
def mk_identity_fn(t):
  if t in _identity_fn_cache:
    return _identity_fn_cache[t]
  f, b, (x,) = build_fn([t], t, name = "ident")
  b.return_(x)
  _identity_fn_cache[t] = f
  return f

_cast_fn_cache = {}
def mk_cast_fn(from_type, to_type):
  key = (from_type, to_type)
  if key in _cast_fn_cache:
    return _cast_fn_cache[key]
  f, b, (x,) = build_fn([from_type], to_type, name = "cast_%s_%s" % (from_type, to_type))
  b.return_(b.cast(x, to_type))
  _cast_fn_cache[key] = f
  return f

_prim_fn_cache = {}
def mk_prim_fn(prim, arg_types):
  key = prim, tuple(arg_types)
  if key in _prim_fn_cache:
    return _prim_fn_cache[key]
  
  upcast_types = prim.expected_input_types(arg_types)
  result_type = prim.result_type(upcast_types)
  f, b, args = build_fn(arg_types, result_type, name =  prim.name)
  # builder will auto-cast argument vars to appropriate types for the given primitive
  b.return_(b.prim(prim, args))
  _prim_fn_cache[key] = f 
  return f 
  



from parakeet import jit, config 
import numpy as np 

def dist(x, y, z, L, periodicX, periodicY, periodicZ):
    " ""Computes distances between all particles and places the result in a matrix such that the ij th matrix entry corresponds to the distance between particle i and j"" "
    N = len(x)
    xtemp = np.tile(x,(N,1))
    dx = xtemp - xtemp.T
    ytemp = np.tile(y,(N,1))
    dy = ytemp - ytemp.T
    ztemp = np.tile(z,(N,1))
    dz = ztemp - ztemp.T

    # Particles 'feel' each other across the periodic boundaries
    if periodicX:
        dx[dx>L/2]=dx[dx > L/2]-L
        dx[dx<-L/2]=dx[dx < -L/2]+L
    if periodicY:
        dy[dy>L/2]=dy[dy>L/2]-L
        dy[dy<-L/2]=dy[dy<-L/2]+L
    if periodicZ:
        dz[dz>L/2]=dz[dz>L/2]-L
        dz[dz<-L/2]=dz[dz<-L/2]+L

    # Total Distances
    d = np.sqrt(dx**2+dy**2+dz**2)

    # Mark zero entries with negative 1 to avoid divergences
    d[d==0] = -1

    return d, dx, dy, dz

@jit 
def parakeet_dist(x, y, z, L, periodicX, periodicY, periodicZ):
  N = len(x)
  def periodic_diff(x1, x2, periodic):
    diff = x1 - x2 
    if periodic:
      if diff > (L / 2):
        diff -= L
      if diff < (-L/2):
        diff += L
    return diff
  dx = np.array([[periodic_diff(x1, x2, periodicX) for x1 in x] for x2 in x])
  dy = np.array([[periodic_diff(y1, y2, periodicY) for y1 in y] for y2 in y])
  dz = np.array([[periodic_diff(z1, z2, periodicZ) for z1 in z] for z2 in z])
  d = dx**2 + dy**2 + dz**2
  for i in xrange(N):
    for j in xrange(N):
      if d[i,j] == 0: d[i,j] = -1 
      else: d[i,j] = np.sqrt(d[i,j])
  return d, dx, dy, dz 

def periodic_diff(x1, x2, L, periodic):
  diff = x1 - x2 
  if periodic:
    if diff > (L / 2):
      diff -= L
    if diff < (-L/2):
      diff += L
  return diff

@jit 
def loopy_dist(x, y, z, L, periodicX, periodicY, periodicZ):
  N = len(x)
  dx = np.zeros((N,N))
  dy = np.zeros( (N,N) )
  dz = np.zeros( (N,N) )
  d = np.zeros( (N,N) )
  for i in xrange(N):
    for j in xrange(N):
      dx[i,j] = periodic_diff(x[j], x[i], L, periodicX)
      dy[i,j] = periodic_diff(y[j], y[i], L, periodicY)
      dz[i,j] = periodic_diff(z[j], z[i], L, periodicZ)
      d[i,j] = dx[i,j] ** 2 + dy[i,j] ** 2 + dz[i,j] ** 2 
      if d[i,j] == 0:
          d[i,j] = -1
      else:
          d[i,j] = np.sqrt(d[i,j])
  return d, dx, dy, dz 
from timer import timer

N = 2000 
x = y = z = np.random.rand(N)
L = 4
periodic = True
with timer("Python #1"):
  d, dx, dy, dz = dist(x, x, x, L,periodic, periodic, periodic)

with timer("Python #2"):
  d, dx, dy, dz = dist(x, x, x, L, periodic, periodic, periodic)

with timer("Parakeet Dist #1", suppress_stdout = False, suppress_stderr = False):
  pd, pdx, pdy, pdz = parakeet_dist(x,x,x,L, periodic, periodic, periodic)

assert np.allclose(pdx, dx), (pdx-dx, np.max(np.abs(pdx - dx)))
assert np.allclose(pdy, dy), np.max(np.abs(pdy - dy))
assert np.allclose(pdz, dz), np.max(np.abs(pdz - dz))
assert np.allclose(pd, d), (np.max(np.abs(pd - d)), pd -d)


with timer("Parakeet Dist #2"):
  pd, pdx, pdy, pdz = parakeet_dist(x,x,x,L, periodic, periodic, periodic)


with timer("Parakeet Loop Dist #1"):
  ld, ldx, ldy, ldz = loopy_dist(x,x,x,L, periodic, periodic, periodic)

assert np.allclose(ldx, dx), (ldx-dx, np.max(np.abs(ldx - dx)))
assert np.allclose(ldy, dy), np.max(np.abs(ldy - dy))
assert np.allclose(ldz, dz), np.max(np.abs(ldz - dz))
assert np.allclose(ld, d), (np.max(np.abs(ld - d)), ld -d)

with timer("Parakeet Loop Dist #2"):
  ld, ldx, ldy, ldz = loopy_dist(x,x,x,L, periodic, periodic, periodic)




from collections import namedtuple
import numpy as np 

from .. import names, prims  
from ..ndtypes import (IntT, FloatT, TupleT, FnT, Type, BoolT, NoneT, Float32, Float64, Bool, 
                       ClosureT, ScalarT, PtrT, NoneType, ArrayT, SliceT, TypeValueT)    
from ..syntax import (Const, Var,  PrimCall, Attribute, TupleProj, Tuple, ArrayView,
                      Expr, Closure, TypedFn)
# from ..syntax.helpers import get_types   
import type_mappings
from base_compiler import BaseCompiler


CompiledFlatFn = namedtuple("CompiledFlatFn", 
                            ("name", "sig", "src",
                             "extra_objects",
                             "extra_functions",
                             "extra_function_signatures", 
                             "declarations"))


# mapping from (field_types, struct_name, field_names) to type names 
_struct_type_names = {}

# mapping from struct name to decl 
_struct_type_decls = {}

class FnCompiler(BaseCompiler):
  
  def __init__(self,  
               module_entry = False, 
               struct_type_cache = None, 
               **kwargs):
    BaseCompiler.__init__(self, **kwargs)
    
    
    self.declarations = []
    
    # depends on these .o files
    self.extra_objects = set([]) 
    
    # to avoid adding the same function's source twice 
    # we use its signature as a key  
    self.extra_functions = {}
    self.extra_function_signatures = []
    
      
    # are we actually compiling the entry-point into a Python module?
    # if so, expect some of the methods like visit_Return to be overloaded 
    # to return PyObjects
    self.module_entry = module_entry
     
  def add_decl(self, decl):
    if decl not in self.declarations:
      self.declarations.append(decl)
  
  def ptr_struct_type(self, elt_t):
    # need both an actual data pointer 
    # and an optional PyObject base
    field_types = ["%s*" % self.to_ctype(elt_t), "PyObject*"]
    return self.struct_type_from_fields(field_types, 
                                        struct_name = "%s_ptr_type" % elt_t,
                                        field_names = ["raw_ptr", "base"])
  
  def array_struct_type(self, elt_t, rank):
    ptr_field_t = self.ptr_struct_type(elt_t)
    field_types = [ptr_field_t, "npy_intp", "npy_intp", "int64_t", "int64_t"]
    field_names = ["data", "shape", "strides", "offset", "size"]
    field_repeats = {}
    field_repeats["shape"] = rank 
    field_repeats["strides"] = rank
    return self.struct_type_from_fields(field_types, "array_type", field_names, field_repeats)
  
  def slice_struct_type(self, start_t = "int64_t", stop_t = "int64_t", step_t = "int64_t"):
    field_types = [start_t, stop_t, step_t]
    field_names = ["start", "stop", "step"]
    return self.struct_type_from_fields(field_types, "slice_type", field_names)
       
  def struct_type_from_fields(self, 
                                 field_types, 
                                 struct_name = "tuple_type", 
                                 field_names = None, 
                                 field_repeats = {}):
    
    if any(not isinstance(t, str) for t in field_types):
      field_types = tuple(self.to_ctype(t) if isinstance(t, Type) else t 
                          for t in field_types)
    else:
      field_types = tuple(field_types)
    
    if field_names is None:
      field_names = tuple("elt%d" % i for i in xrange(len(field_types)))
    else:
      assert len(field_names) == len(field_types), \
        "Mismatching number of types %d and field names %d" % (len(field_types), len(field_names))
      field_names = tuple(field_names)
    
    repeat_set =  frozenset(sorted(field_repeats.items()))
    key = field_types, struct_name, field_names, repeat_set
     
    if key in _struct_type_names:
      typename = _struct_type_names[key]
      decl = _struct_type_decls[typename]
      if decl not in self.declarations:
        self.declarations.append(decl)
      return typename 
    
    typename = names.fresh(struct_name).replace(".", "")
    
    field_decls = []
    for t, field_name in zip(field_types, field_names):
      if field_name in field_repeats:
        field_decl = "  %s %s[%d];" % (t, field_name, field_repeats[field_name])
      else:
        field_decl = "  %s %s;" % (t, field_name)
      field_decls.append(field_decl)
    decl = "typedef struct %s {\n%s\n} %s;" % (typename, "\n".join(field_decls), typename)
    
    _struct_type_names[key] = typename
    _struct_type_decls[typename] = decl
    self.add_decl(decl)
    return typename 
  

  def to_ctype(self, parakeet_type):
    if isinstance(parakeet_type, (NoneT, ScalarT)):
      return type_mappings.to_ctype(parakeet_type)
    
    elif isinstance(parakeet_type, TupleT):
      return self.struct_type_from_fields(parakeet_type.elt_types)
    elif isinstance(parakeet_type, PtrT):
      return self.ptr_struct_type(parakeet_type.elt_type)
    elif isinstance(parakeet_type, ArrayT):
      elt_t = parakeet_type.elt_type 
      rank = parakeet_type.rank 
      return self.array_struct_type(elt_t, rank)
    
    elif isinstance(parakeet_type, SliceT):
      return self.slice_struct_type()
    elif isinstance(parakeet_type, ClosureT):
      return self.struct_type_from_fields(parakeet_type.arg_types)
    elif isinstance(parakeet_type, TypeValueT):
      return "int"
    else:
      assert False, "Don't know how to make C type for %s" % parakeet_type
    
  
  def to_ctypes(self, ts):
    return tuple(self.to_ctype(t) for t in ts)
  
  
  
  def visit_Slice(self, expr):
    typename = self.to_ctype(expr.type)
    start = self.visit_expr(expr.start)
    stop = self.visit_expr(expr.stop)
    step = self.visit_expr(expr.step)
    return self.fresh_var(typename, "slice", "{%s, %s, %s}" % (start,stop,step))
    
  def visit_Alloc(self, expr):
    elt_t =  expr.elt_type
    nelts = self.fresh_var("npy_intp", "nelts", self.visit_expr(expr.count))
    bytes_per_elt = elt_t.nbytes
    nbytes = self.mul(nelts, bytes_per_elt)#"%s * %d" % (nelts, bytes_per_elt)
    raw_ptr = "(%s) malloc(%s)" % (type_mappings.to_ctype(expr.type), nbytes)
    struct_type = self.to_ctype(expr.type)
    return self.fresh_var(struct_type, "new_ptr", "{%s, NULL}" % raw_ptr)
    
  def visit_Const(self, expr):
    t = expr.type 
    c = t.__class__ 
    if c == BoolT:
      return "1" if expr.value else "0"
    elif c == NoneT:
      return "0"
    
    assert isinstance(t, ScalarT), "Don't know how to translate Const %s : %s" % (expr,t)
    v = expr.value 
    if np.isinf(v):
      return "INFINITY"
    elif np.isnan(v):
      return "NAN"
    return "%s" % expr.value 
  
  def visit_Var(self, expr):
    return self.name(expr.name)
  
  def visit_Cast(self, expr):
    x = self.visit_expr(expr.value)
    ct = self.to_ctype(expr.type)
    if isinstance(expr, (Const, Var)):
      return "(%s) %s" % (ct, x)
    else:
      return "((%s) (%s))" % (ct, x)
  
  

  
  
  def visit_PrimCall(self, expr):
    t = expr.type
    args = self.visit_expr_list(expr.args)
    
    # parenthesize any compound expressions 
    for i, arg_expr in enumerate(expr.args):
      if not isinstance(arg_expr, (Var, Const)):
        args[i] = "(" + args[i] + ")"
        
    p = expr.prim 
    if p == prims.add:
      #return "%s + %s" % (args[0], args[1])
      return self.add(args[0], args[1])
    if p == prims.subtract:
      #return "%s - %s" % (args[0], args[1])
      return self.sub(args[0],args[1])
    elif p == prims.multiply:
      return self.mul(args[0], args[1])
      # return "%s * %s" % (args[0], args[1])
    elif p == prims.divide:
      return self.div(args[0], args[1])
      # return "%s / %s" % (args[0], args[1])
    elif p == prims.negative:
      if t == Bool:
        return "1 - %s" % args[0]
      else:
        return "-%s" % args[0]
    elif p == prims.abs:
      x  = args[0]
      return " %s >= 0 ? %s  : -%s" % (x,x,x)
    
    elif p == prims.bitwise_and:
      return "%s & %s" % (args[0], args[1])
    elif p == prims.bitwise_or:
      return "%s | %s" % (args[0], args[1])
    elif p == prims.bitwise_not:
      return "~%s" % args[0]
    
    elif p == prims.logical_and:
      return self.and_(args[0], args[1])
      
    elif p == prims.logical_or:
      return self.or_(args[0], args[1])
    
    elif p == prims.logical_not:
      return self.not_(args[0])
      
    elif p == prims.equal:
      return self.eq(args[0], args[1], t)
    
    elif p == prims.not_equal:
      return self.neq(args[0], args[1], t)
    
    elif p == prims.greater:
      return self.gt(args[0], args[1], t)
      
    elif p == prims.greater_equal:
      return self.gte(args[0], args[1], t)
    
    elif p == prims.less:
      return self.lt(args[0], args[1], t)
    
    elif p == prims.less_equal:
      return self.lte(args[0], args[1], t)
    
    elif p == prims.remainder:
      x,y = args
      if t == Float32: return "fmod(%s, %s)" % (x,y)
      elif t == Float64: return "fmod(%s, %s)" % (x,y)
      assert isinstance(t, (BoolT, IntT)), "Modulo not implemented for %s" % t
      rem = self.fresh_var(t, "rem", "%s %% %s" % (x,y))
      y_is_negative = self.fresh_var(t, "y_is_negative", "%s < 0" % y)
      rem_is_negative = self.fresh_var(t, "rem_is_negative", "%s < 0" % rem)
      y_nonzero = self.fresh_var(t, "y_nonzero", "%s != 0" % y)
      rem_nonzero = self.fresh_var(t, "rem_nonzero", "%s != 0" % rem)
      neither_zero = self.fresh_var(t, "neither_zero", "%s && %s" % (y_nonzero, rem_nonzero))
      diff_signs = self.fresh_var(t, "diff_signs", "%s ^ %s" % (y_is_negative, rem_is_negative))
      should_flip = self.fresh_var(t, "should_flip", "%s && %s" % (neither_zero, diff_signs))
      flipped_rem = self.fresh_var(t, "flipped_rem", "%s + %s" % (y, rem))
      return "%s ? %s : %s" % (should_flip, flipped_rem, rem)
    elif p == prims.fmod:
      if t == Float32: return "fmodf(%s, %s)" % (args[0], args[1])
      elif t == Float64: return "fmod(%s, %s)" % (args[0], args[1])
      return "%s %% %s" % (args[0], args[1])
    elif p == prims.maximum:
      x,y = args
      return "(%s > %s) ? %s : %s" % (x,y,x,y)
    elif p == prims.minimum:
      x,y = args
      return "(%s < %s) ? %s : %s" % (x,y,x,y)
    
    elif p == prims.power:
      if t == Float32: 
        return "powf(%s, %s)" % (args[0], args[1])
      else:
        return "pow(%s, %s)" % (args[0], args[1])
    
    elif isinstance(t, FloatT):
      # many float prims implemented using the same name in math.h
      name = p.name
      if name.startswith("arc"):
        # arccos -> acos
        name = "a" + name[3:]
      if t == Float32: name = name + "f" 
      if len(args) == 1:
        return "%s(%s)" % (name, args[0])
      else:
        assert len(args) == 2, "Unexpected prim %s with %d args (%s)" % (p, len(args), args)
        return "%s(%s, %s)" % (name, args[0], args[1])
  
    else:
      assert False, "Prim not yet implemented: %s" % p
  
  def visit_Index(self, expr):
    arr = self.visit_expr(expr.value)
    if isinstance(expr.index.type, ScalarT):
      index_exprs = [expr.index]
    else:
      assert isinstance(expr.index.type, TupleT), \
        "Unexpected index %s : %s" % (expr.index, expr.index.type)
      if isinstance(expr.index, Tuple):
        index_exprs = expr.index.elts 
      else:
        index_exprs = [TupleProj(expr.index, i, type = t) 
                       for i, t in enumerate(expr.index.type.elt_types) ]
    assert all(isinstance(idx_expr.type, ScalarT) for idx_expr in index_exprs), \
      "Expected all indices to be scalars but got %s" % (index_exprs,)
    indices = [self.visit_expr(idx_expr) for idx_expr in index_exprs]
    if isinstance(expr.value.type, PtrT):
      assert len (indices) == 1, \
        "Can't index into pointer using %d indices (%s)" % (len(indices), index_exprs)
      raw_ptr = "%s.raw_ptr" % arr
      offset = indices[0]
    else:
      assert isinstance(expr.value.type, ArrayT)
      offset = self.fresh_var("int64_t", "offset", "%s.offset" % arr)
      for i, idx in enumerate(indices):
        stride = "%s.strides[%d]" % (arr, i)
        self.append("%s += %s * %s;" % (offset, idx, stride))
      raw_ptr = "%s.data.raw_ptr" % arr 

    return "%s[%s]" % (raw_ptr, offset)
  
  def visit_Call(self, expr):
    fn_name = self.get_fn_name(expr.fn)
    closure_args = self.get_closure_args(expr.fn)
    args = self.visit_expr_list(expr.args)
    return "%s(%s)" % (fn_name, ", ".join(tuple(closure_args) + tuple(args)))
  
  def visit_Select(self, expr):
    cond = self.visit_expr(expr.cond)
    true = self.visit_expr(expr.true_value)
    false = self.visit_expr(expr.false_value)
    return "%s ? %s : %s" % (cond, true, false) 
  
  def is_pure(self, expr):
    return expr.__class__ in (Var, Const, PrimCall, Attribute, TupleProj, Tuple, ArrayView)
  
  def visit_Assign(self, stmt):
    rhs = self.visit_expr(stmt.rhs)

    if stmt.lhs.__class__ is Var:
      lhs = self.visit_expr(stmt.lhs)
      return "%s %s = %s;" % (self.to_ctype(stmt.lhs.type), lhs, rhs)
    elif stmt.lhs.__class__ is Tuple:
      struct_value = self.fresh_var(self.to_ctype(stmt.lhs.type), "lhs_tuple")
      self.assign(struct_value, rhs)
      
      for i, lhs_var in enumerate(stmt.lhs.elts):
        assert isinstance(lhs_var, Var), "Expected LHS variable, got %s" % lhs_var
        c_name = self.visit_expr(lhs_var)
        self.append("%s %s = %s.elt%d;" % (self.to_ctype(lhs_var.type), c_name, struct_value, i ))
      return "" 
    else:
      lhs = self.visit_expr(stmt.lhs)
      return "%s = %s;" % (lhs, rhs)
  
  def declare(self, parakeet_name, parakeet_type, init_value = None):
    c_name = self.name(parakeet_name)
    t = self.to_ctype(parakeet_type)
    if init_value is None:
      self.append("%s %s;" % (t, c_name))
    else: 
      self.append("%s %s = %s;" % (t, c_name, init_value))
  
  def declare_merge_vars(self, merge):
    """ 
    Declare but don't initialize
    """
    for (name, (left, _)) in merge.iteritems():
      self.declare(name, left.type)
      
  def visit_merge_left(self, merge, fresh_vars = True):
    
    if len(merge) == 0:
      return ""
    
    self.push()
    self.comment("Merge Phi Nodes (left side) " + str(merge))
    for (name, (left, _)) in merge.iteritems():
      c_left = self.visit_expr(left)
      if fresh_vars:
        self.declare(name, left.type, c_left)
      else:
        c_name = self.name(name)
        self.append("%s = %s;" % (c_name, c_left))
        
    return self.pop()
  
  def visit_merge_right(self, merge):
    
    if len(merge) == 0:
      return ""
    self.push()
    self.comment("Merge Phi Nodes (right side) " + str(merge))
    
    for (name, (_, right)) in merge.iteritems():
      c_right = self.visit_expr(right)
     
      self.append("%s = %s;"  % (self.name(name), c_right))
    return self.pop()
  
  def visit_NumCores(self, expr):
    # by default we're running sequentially 
    return "1"
  
  def visit_Comment(self, stmt):
    return "// " + stmt.text
    
  def visit_PrintString(self, stmt):
    self.printf(stmt.text)
    return "// done with printf"
    
  def visit_SourceExpr(self, expr):
    return expr.text 
  
  def visit_SourceStmt(self, stmt):
    return stmt.text 
  
  def visit_If(self, stmt):
    self.declare_merge_vars(stmt.merge)
    cond = self.visit_expr(stmt.cond)
    true = self.visit_block(stmt.true) + self.visit_merge_left(stmt.merge, fresh_vars = False)
    false = self.visit_block(stmt.false) + self.visit_merge_right(stmt.merge)
    return self.indent("if(%s) {\n%s\n} else {\n%s\n}" % (cond, self.indent(true), self.indent(false))) 
  
  def visit_While(self, stmt):
    decls = self.visit_merge_left(stmt.merge, fresh_vars = True)
    cond = self.visit_expr(stmt.cond)
    body = self.visit_block(stmt.body) + self.visit_merge_right(stmt.merge)
    return decls + "while (%s) {%s}" % (cond, body)
  
  def visit_ExprStmt(self, stmt):
    return self.visit_expr(stmt.value) + ";"
  
  def visit_ForLoop(self, stmt):
    s = self.visit_merge_left(stmt.merge, fresh_vars = True)
    start = self.visit_expr(stmt.start)
    stop = self.visit_expr(stmt.stop)
    step = self.visit_expr(stmt.step)
    var = self.visit_expr(stmt.var)
    t = self.to_ctype(stmt.var.type)
    body =  self.visit_block(stmt.body)
    body += self.visit_merge_right(stmt.merge)
    body = self.indent("\n" + body) 
    s += "\n %(t)s %(var)s;"
    up_loop = \
        "\nfor (%(var)s = %(start)s; %(var)s < %(stop)s; %(var)s += %(step)s) {%(body)s}"
    down_loop = \
        "\nfor (%(var)s = %(start)s; %(var)s > %(stop)s; %(var)s += %(step)s) {%(body)s}"
      
    if stmt.step.__class__ is Const:
      if stmt.step.value >= 0:
        s += up_loop
      else:
        s += down_loop
    else:
      s += "if(%(step)s >= 0) {\n"
      s += up_loop
      s += "\n} else {\n"
      s += down_loop
      s += "\n}"
    return s % locals()

  def visit_Return(self, stmt):
    assert not self.return_by_ref, "Returning multiple values by ref not yet implemented: %s" % stmt
    if self.return_void:
      return "return;"
    elif isinstance(stmt.value, Tuple):
      # if not returning multiple values by reference, then make a struct for them
      struct_type = self.to_ctype(stmt.value.type)
      result_elts = ", ".join(self.visit_expr(elt) for elt in stmt.value.elts)
      result_value = "{" + result_elts + "}"
      result = self.fresh_var(struct_type, "result", result_value)
      return "return %s;" % result 
    else:
      v = self.visit_expr(stmt.value)
      return "return %s;" % v
  

  
  def visit_block(self, stmts, push = True):
    if push: self.push()
    for stmt in stmts:
      s = self.visit_stmt(stmt)
      self.append(s)
    self.append("\n")
    return self.indent("\n" + self.pop())
      
  
  def get_fn_name(self, expr, compiler_kwargs = {}, attributes = [], inline = True):
    if expr.__class__ is  TypedFn:
      fn = expr 
    elif expr.__class__ is Closure:
      fn = expr.fn 
    else:
      assert isinstance(expr.type, (FnT, ClosureT)), \
        "Expected function or closure, got %s : %s" % (expr, expr.type)
      fn = expr.type.fn

    compiler = self.__class__(module_entry = False, **compiler_kwargs)
    compiled = compiler.compile_flat_source(fn, attributes = attributes, inline = inline)
    
    if compiled.sig not in self.extra_function_signatures:
      # add any declarations it depends on 
      for decl in compiled.declarations:
        self.add_decl(decl)
      
      #add any external objects it wants to be linked against 
      self.extra_objects.update(compiled.extra_objects)
      
      # first add the new function's dependencies
      for extra_sig in compiled.extra_function_signatures:
        if extra_sig not in self.extra_function_signatures:
          self.extra_function_signatures.append(extra_sig)
          self.extra_functions[extra_sig] = compiled.extra_functions[extra_sig] 
      # now add the function itself 
      self.extra_function_signatures.append(compiled.sig)
      self.extra_functions[compiled.sig] = compiled.src
      
    
    for link_flag in compiler.extra_link_flags:
      if link_flag not in self.extra_link_flags:
        self.extra_link_flags.append(link_flag)
    
    for compile_flag in compiler.extra_compile_flags:
      if compile_flag not in self.extra_compile_flags:
        self.extra_compile_flags.append(compile_flag)
  
    return compiled.name

  def get_closure_args(self, fn):
    if isinstance(fn.type, FnT):
      return []
    else:
      assert isinstance(fn, Closure), "Expected closure, got %s : %s" % (fn, fn.type)
      return self.visit_expr_list(fn.args)
      
  def build_loops(self, loop_vars, bounds, body):
    if len(loop_vars) == 0:
      return body
    var = loop_vars[0]
    bound = bounds[0]
    nested = self.build_loops(loop_vars[1:], bounds[1:], body)
    return """
    for (%s = 0; %s < %s; ++%s) {
      %s
    }""" % (var, var, bound, var, nested )
  
  def visit_TypedFn(self, expr):
    return self.get_fn_name(expr)

  def visit_UntypedFn(self, expr):
    return "{}"
  
  def return_types(self, fn):
    if isinstance(fn.return_type, TupleT):
      return fn.return_type.elt_types
    elif isinstance(fn.return_type, NoneT):
      return []
    else:
      # assert isinstance(fn.return_type, (PtrT, ScalarT)), "Unexpected return type %s" % fn.return_type
      return [fn.return_type]
    
  
  def visit_flat_fn(self, fn, return_by_ref = False, attributes = None, inline = True):
    if attributes is None:
      attributes = []
    
    c_fn_name = names.refresh(fn.name).replace(".", "_")
    arg_types = [self.to_ctype(t) for t in fn.input_types]
    arg_names = [self.name(old_arg) for old_arg in fn.arg_names]

    return_types = self.return_types(fn)
    n_return = len(return_types)
    
    if n_return == 1:
      return_type = self.to_ctype(return_types[0])
      self.return_void = (return_type == NoneType)
      self.return_by_ref = False
    elif n_return == 0:
      return_type = "void"
      self.return_void = True
      self.return_by_ref = False
    elif return_by_ref:
      return_type = "void"
      self.return_void = True
      self.return_by_ref = True
      self.return_var_types = [self.to_ctype(t) for t in return_types]
      self.return_var_names = [self.fresh_name("return_value%d" % i) for i in xrange(n_return)]
      arg_types = arg_types + ["%s*" % t for t in self.return_var_types] 
      arg_names = arg_names + self.return_var_names
    else:
      return_type = self.struct_type_from_fields(return_types)
      self.return_void = False 
      self.return_by_ref = False 
    args_str = ", ".join("%s %s" % (t, name) for (t,name) in zip(arg_types,arg_names))
    
    body_str = self.visit_block(fn.body) 
    
    if inline:
      # "__attribute__((always_inline))",
      attributes = attributes + ["inline"]
    attr_str = " ".join(attributes)
    sig = "%s %s(%s)" % (return_type, c_fn_name, args_str)
    src = "%s %s {\n\n%s}" % (attr_str, sig, body_str) 
    return c_fn_name, sig, src
  
  @property 
  def cache_key(self):
    """
    If we ever need to differentiate compiled function by *how* they were compiled,
    we can use this cache key to track the class of the compiler or other
    relevant meta-data
    """ 
    return self.__class__ 
  
  _flat_compile_cache = {}
  def compile_flat_source(self, parakeet_fn, attributes = [], inline = True):
      
    # make sure compiled source uses consistent names for tuple and array types, 
    # which both need declarations for their C struct representations  
    struct_types = set(t for t in parakeet_fn.type_env.itervalues() 
                         if isinstance(t, (ArrayT, TupleT)))
    
    # include your own class in the cache key so that we get distinct code 
    # for derived compilers like OpenMP and CUDA 
    key = parakeet_fn.cache_key, frozenset(struct_types), self.cache_key, tuple(attributes)
    
    if key in self._flat_compile_cache:
      return self._flat_compile_cache[key]
    
    name, sig, src = self.visit_flat_fn(parakeet_fn, attributes = attributes, inline = inline)
      
    
    result = CompiledFlatFn(
      name = name, 
      sig = sig, 
      src = src,
      extra_objects = self.extra_objects, 
      extra_functions = self.extra_functions,
      extra_function_signatures = self.extra_function_signatures,
      declarations = self.declarations)
    self._flat_compile_cache[key] = result
    return result



from dsltools import  ScopedDict
 

from .. import prims  
from ..ndtypes import NoneT, NoneType, get_rank 
from ..syntax import Var, Const, PrimCall, Tuple, Slice, TupleProj, Attribute, Shape 
 
import numpy as np 
from ..syntax.helpers import unwrap_constant
from syntax_visitor import SyntaxVisitor



class AbstractValue(object):
  def __eq__(self, other):
    return False    

  def __ne__(self, other):
    return not self == other
  
  def __repr__(self):
    return str(self)
  
  def combine(self, other):
    if self == other:
      return self 
    else:
      return any_value 
  
  def widen(self, other):
    if self == other:
      return self 
    else:
      assert False, "widening not implemented for (%s, %)" % (self, other) 
    
  
class Unknown(AbstractValue):
  """
  Top of the lattice 
  """
  
  def __str__(self):
    return "<unknown>"
  
  def combine(self, other):
    return other
  
  def widen(self, other):
    return other  
  
  
unknown_value = Unknown()

class AnyValue(AbstractValue):
  """
  Bottom of the lattice 
  """
  
  def __str__(self):
    return "<any>"
  
  def combine(self, other):
    return any_value 
  
  def widen(self, other):
    return any_value 
  
  def __eq__(self, other):
    return other.__class__ is AnyValue 
  
any_value = AnyValue()

class NoneValue(AbstractValue):
  def __str__(self):
    return "None"
  
  
  def __eq__(self, other):
    return other.__class__ is NoneValue
  
  def combine(self, other):
    if self == other:
      return self 
    else:
      return any_value 
    
  def widen(self, other):
    return self.combine(other)

class Interval(AbstractValue):
  def __init__(self, lower, upper):
    self.lower = lower 
    self.upper = upper 
  
  def combine(self, other):
    if other.__class__ is not Interval:
      return any_value 
    
    lower = min(self.lower, other.lower)
    upper = max(self.upper, other.upper)
    return Interval(lower,upper)
    
  def __eq__(self, other):
    return other.__class__ is Interval and self.lower == other.lower and self.upper == other.upper
  
  def widen(self, other):
    
    """
    Like 'combine' but not symmetric-- the other value
    is a more recent version from within a loop, 
    so try to follow the 'trend' of its changes. 
    """
    
    if other.__class__ is not Interval:
      return any_value 
    lower_diff = other.lower - self.lower
    upper_diff = other.upper - self.upper
    

    if lower_diff < 0:
      lower = -np.inf
    else:
      lower = self.lower 
    
    
    
    if upper_diff > 0:
      upper = np.inf 
    else:
      upper = self.upper 
    
    if lower == self.lower and upper == self.upper:
      return self 
    else:
      return Interval(lower, upper)     
    
  def __str__(self):
    return "[%s,%s]" % (self.lower, self.upper)
    
const_zero = Interval(0,0)
const_one = Interval(1,1)
const_none = NoneValue()
positive_interval = Interval(0, np.inf)
negative_interval = Interval(-np.inf, 0)
  
class TupleOfIntervals(AbstractValue):
  def __init__(self, elts):
    assert all(isinstance(elt, AbstractValue) for elt in elts)
    self.elts = elts 
    
  def combine(self, other):
    if other.__class__ is not TupleOfIntervals:
      return any_value 
    elif len(other.elts) != len(self.elts):
      return any_value  
    combined_elts = [e1.combine(e2) for e1,e2 in zip(self.elts, other.elts)]
    return TupleOfIntervals(combined_elts)

  def widen(self, other):
    if other.__class__ is not TupleOfIntervals or len(other.elts) != len(self.elts):
      return any_value 
    return mk_tuple([e1.widen(e2) for e1,e2 in zip(self.elts, other.elts)])
  
  def __str__(self):
    return "tuple(%s)" % ", ".join(str(elt) for elt in self.elts)

  def __eq__(self, other):
    return other.__class__ is TupleOfIntervals and len(other.elts) == len(self.elts) and \
      all(e1 == e2 for e1,e2 in zip(self.elts,other.elts))

def mk_tuple(elts):
  if all(elt is None or elt is any_value for elt in elts):
    return any_value 
  else:
    return TupleOfIntervals(elts)  

class SliceOfIntervals(AbstractValue):
  def __init__(self, start, stop, step):
    self.start = start 
    self.stop = stop 
    self.step = step 
    
  def combine(self, other):
    if other.__class__ is not SliceOfIntervals:
      return any_value
    start = self.start.combine(other.start)
    stop = self.stop.combine(other.stop)
    step = self.step.combine(other.step)
    if start == self.start and stop == self.stop and step == self.stop:
      return self 
    return mk_slice(start,stop,step)

  def __eq__(self, other):
    return other.__class__ is SliceOfIntervals and \
      self.start == other.start and \
      self.stop == other.stop and \
      self.step == other.step

  def widen(self, other):
    if other.__class__ is not SliceOfIntervals:
      return any_value 
    start = self.start.widen(other.start)
    stop = self.stop.widen(other.stop)
    step = self.step.widen(other.step)
    if start == self.start and stop == self.stop and step == self.step:
      return self 
    return mk_slice(start,stop,step)
        
  def __str__(self):
    return "slice(start=%s, stop=%s, step=%s)" % (self.start, self.stop, self.step)
    
def mk_slice(start, stop, step):
  if (start is None or start is any_value) and \
     (stop is None or stop is any_value)  and \
     (step is None or step is any_value):
    return any_value 
  return SliceOfIntervals(start, stop, step) 



class ValueRangeAnalyis(SyntaxVisitor):
  def __init__(self):
    SyntaxVisitor.__init__(self)
    self.ranges = {} 
    self.old_values = ScopedDict()
    self.old_values.push()

  def get(self, expr):
    c = expr.__class__ 
    
    if expr.type.__class__ is NoneT:
      return const_none 
    elif c is Const:
      return Interval(expr.value, expr.value)
    
    elif c is Var and expr.name in self.ranges:
      return self.ranges[expr.name]
    
    elif c is Tuple:
      elt_values = [self.get(elt) for elt in expr.elts]
      return mk_tuple(elt_values)
    
    elif c is TupleProj:
      tup = self.get(expr.tuple)
      idx = unwrap_constant(expr.index) 
      if tup.__class__ is TupleOfIntervals:
        return tup.elts[idx]
      
    elif c is Slice:
      start = self.get(expr.start)
      stop = self.get(expr.stop)
      if expr.step.type ==  NoneType:
        step = const_one   
      else:
        step = self.get(expr.step)
      return mk_slice(start, stop, step)
    
    elif c is Shape:
      ndims = get_rank(expr.array.type)   
      return mk_tuple([positive_interval] * ndims)
    
    elif c is Attribute:
      if expr.name == 'shape':
        ndims = get_rank(expr.value.type)   
        return mk_tuple([positive_interval] * ndims)
      elif expr.name == 'start':
        sliceval = self.get(expr.value)
        if isinstance(sliceval, SliceOfIntervals):
          return sliceval.start
      elif expr.name == 'stop':
        sliceval = self.get(expr.value)
        if isinstance(sliceval, SliceOfIntervals):
          return sliceval.stop
      elif expr.name == 'step':
        sliceval = self.get(expr.value)
        if isinstance(sliceval, SliceOfIntervals):
          return sliceval.step
    elif c is PrimCall:
      p = expr.prim 
      if p.nin == 2:
        x = self.get(expr.args[0])
        y = self.get(expr.args[1])
        if p == prims.add:

          return self.add_range(x,y)
        elif p == prims.subtract:
          return self.sub_range(x,y)
        elif p == prims.multiply:
          return self.mul_range(x,y)
      elif p.nin == 1:
        x = self.get(expr.args[0])
        if p == prims.negative:
          return self.neg_range(x)
    return any_value 
  
  def set(self, name, val):
    if val is not None and val is not unknown_value:
      old_value = self.ranges.get(name, unknown_value)
      self.ranges[name] = val
      if old_value != val and old_value is not unknown_value:
        self.old_values[name] = old_value 

      
  def add_range(self, x, y):
    if not isinstance(x, Interval) or not isinstance(y, Interval):
      return any_value 
    return Interval (x.lower + y.lower, x.upper + y.upper)
  
  def sub_range(self, x, y):
    if not isinstance(x, Interval) or not isinstance(y, Interval):
      return any_value
    return Interval(x.lower - y.upper, x.upper - y.lower)
  
  def mul_range(self, x, y):
    if not isinstance(x, Interval) or not isinstance(y, Interval):
      return any_value
    xl, xu = x.lower, x.upper 
    yl, yu = y.lower, y.upper 
    products = (xl * yl, xl * yu, xu * yl, xu * yu)
    lower = min(products)
    upper = max(products)
    return Interval(lower, upper)
  
  def neg_range(self, x):
    if not isinstance(x, Interval):
      return any_value 
    return Interval(-x.upper, -x.lower) 
  
  def visit_Assign(self, stmt):
    if stmt.lhs.__class__ is Var:
      name = stmt.lhs.name 
      v = self.get(stmt.rhs)
      self.set(name, v)
  
  def visit_merge_left(self, phi_nodes):
    for (k, (left, _)) in phi_nodes.iteritems():
      left_val = self.get(left)
      self.set(k, left_val)
      
      
  def visit_merge(self, phi_nodes):    
    for (k, (left,right)) in phi_nodes.iteritems():
      left_val = self.get(left)
      right_val = self.get(right)
      self.set(k, left_val.combine(right_val))

  def visit_Select(self, expr):
    return self.get(expr.true_value).combine(self.get(expr.false_value))
  
  def always_positive(self, x, inclusive = True):
    if not isinstance(x, Interval):
      return False 
    elif inclusive:
      return x.lower >= 0
    else:
      return x.lower > 0
  
  def always_negative(self, x, inclusive = True):
    if not isinstance(x, Interval):
      return False 
    elif inclusive:
      return x.upper <= 0
    else:
      return x.upper < 0
    
  
  def widen(self, old_values):
    for (k, oldv) in old_values.iteritems():
      newv = self.ranges[k]
      if oldv != newv:
        self.ranges[k] = oldv.widen(newv)
          
  def run_loop(self, body, merge):
    
    
    # run loop for the first time 
    self.old_values.push()
    self.visit_merge_left(merge)
    self.visit_block(body)
    self.visit_merge(merge)
    
    #run loop for the second time 
    self.visit_block(body)
    self.visit_merge(merge)
    
    old_values = self.old_values.pop()
    self.widen(old_values)
    
    # TODO: verify that it's safe not to run loop with widened values
    #self.visit_block(body)
    #self.visit_merge(merge)
    
    
  def visit_While(self, stmt):
    self.run_loop(stmt.body, stmt.merge)
  
  
  def visit_ForLoop(self, stmt):    

    start = self.get(stmt.start)
    stop = self.get(stmt.stop)
    step = self.get(stmt.step)
    name = stmt.var.name
    iterator_range = any_value  
    if isinstance(start, Interval) and isinstance(stop, Interval):
      lower = min (start.lower, start.upper, stop.lower, stop.upper)
      upper = max (start.lower, start.upper, stop.lower, stop.upper)
      iterator_range = Interval(lower,upper)
      
    elif isinstance(start, Interval) and isinstance(step, Interval):
      if self.always_positive(step):
        iterator_range = Interval(start.lower, np.inf)
      elif self.always_negative(step, inclusive = False):
        iterator_range = Interval(-np.inf, start.upper)
    
    elif isinstance(stop, Interval) and isinstance(step, Interval):
      if self.always_positive(step):
        iterator_range = Interval(-np.inf, stop.upper)
      elif self.always_negative(step, inclusive = False):
        iterator_range = Interval(stop.lower, np.inf)  
    self.set(name, iterator_range)
    self.run_loop(stmt.body, stmt.merge)

      
    


import numpy as np
import parakeet
from parakeet.testing_helpers import expect, run_local_tests, expect_each


m = 5
n = 3
float_mat = np.random.random((m,n))
int_mat = float_mat.astype('int16')
bool_mat = int_mat > 0
matrices = [float_mat, int_mat, bool_mat]

def each_sum(X):
  return parakeet.each(np.sum, X)

def test_each_sum():
  expect_each(each_sum, lambda X: np.sum(X, axis=1), matrices)

def sum_cols(X):
  return np.sum(X, axis = 0)

def test_col_sum():
  expect_each(sum_cols, lambda X: np.sum(X, axis=0), matrices)

def sum_rows(X):
  return np.sum(X, axis = 1)

def test_sum_rows():
  expect_each(sum_rows, lambda X: np.sum(X, axis=1), matrices)

def test_sum_elts():
  expect_each(np.sum, np.sum, matrices)

if __name__ == '__main__':
  run_local_tests()



import numpy as np

from parakeet import each
from parakeet.testing_helpers import (run_local_tests, expect, expect_each, eq, expect_allpairs)

ints_1d = np.arange(300, dtype='int')
floats_1d = np.arange(300, dtype='float')

ints_2d = np.reshape(ints_1d, (25,12))
floats_2d = np.reshape(ints_2d, (25,12))

bools_1d = ints_1d % 2 == 0
bools_2d = ints_2d % 2 == 0

vecs = [ints_1d, floats_1d, bools_1d]
matrices = [ints_2d, floats_2d, bools_2d]
all_arrays = vecs + matrices

def add1_scalar(x):
  return x+1

def test_add1_external_map():
  parakeet_result = each(add1_scalar, ints_1d)
  python_result = ints_1d + 1
  assert eq(parakeet_result, python_result), \
         "Python %s != Parakeet %s" % (python_result, parakeet_result)

def add1_map(x_vec):
  return each(add1_scalar, x_vec)

def test_add1_internal_map_vecs():
  expect_each(add1_map, add1_scalar, vecs)

def test_add1_internal_map_matrices():
  expect_each(add1_map, add1_scalar, matrices)

def add(x,y):
  return x + y

def test_implicit_add_vec():
  expect_allpairs(add, np.add, vecs)

def test_implicit_add_mat():
  expect_allpairs(add, np.add, matrices)

def each_add(x,y):
  return each(add, x, y)

def test_explicit_add_vec():
  expect_allpairs(each_add, np.add, vecs)

def test_explicit_add_mat():
  expect_allpairs(each_add, np.add, matrices)

def conditional_div(x,y):
  if y == 0:
    return 0
  else:
    return x / y

def python_conditional_div(x,y):
  result = [xi / yi if yi != 0 else 0 for (xi,yi) in zip(x,y)]
  return np.array(result)

def each_conditional_div(x,y):
  return each(conditional_div, x, y)

def test_conditional_div():
    expect_allpairs(each_conditional_div, python_conditional_div,
                    [ints_1d, floats_1d])

def second_elt(x):
  return x[1]

X = np.array([[1,2,3],[4,5,6]])

def second_of_columns(X):
  return each(second_elt, X, axis=1)

def test_second_of_columns():
  expect(second_of_columns, [X], np.array([4,5,6]))

def second_of_rows(X):
  return each(second_elt, X, axis=0)

def test_second_of_rows():
  expect(second_of_rows, [X], np.array([2,5]))

def nested_each(x):
  def dummy(x):
    def dummy2():
      return x
    return dummy2()
  return each(dummy, x)

def test_nested_each():
  expect(nested_each, [X], X)

def ident(x):
  return x

def each_ident(x):
  return each(ident, x)

def each_each_ident(x):
  return each(each_ident, x)

def test_map_map():
  expect(each_each_ident, [X], X)

if __name__ == '__main__':
  run_local_tests()



import numpy as np
from parakeet.testing_helpers import  expect_allpairs, run_local_tests

bool_vec = np.array([True, False, True, ])
int_vec = np.array([1,2,3,])
float_vec = np.array([10.0, 20.0, 30.0 ])

vectors = [bool_vec, int_vec, float_vec]

def loop_dot(x,y):
  n = x.shape[0]
  result = x[0] * y[0]
  i = 1
  while i < n:
    result += x[i] * y[i]
    i = i + 1
  return result

def test_loopdot():
  expect_allpairs(loop_dot, np.dot, vectors)

def dot(x,y):
  return sum(x*y)

def test_adverb_dot():
  expect_allpairs(dot, lambda x,y: np.sum(x*y), vectors)

if __name__ == '__main__':
  run_local_tests()



import numpy as np 
import parakeet
from parakeet.testing_helpers import run_local_tests 


@parakeet.jit
def rint(x):
    return np.rint(x)

def test_rint():
    assert np.rint(1.2) == rint(1.2)
    assert np.rint(-1) == rint(-1)
    assert np.rint(-0.1) == rint(-0.1)
    x = np.array([-1, 2, 0])
    assert np.allclose(np.rint(x), rint(x))
    y = np.array([-1.1, -0.6, -0.4, 0.4, 0.6])
    assert np.allclose(np.rint(y), rint(y))

if __name__ == "__main__":
    run_local_tests()



import numpy as np

import dtypes 
import type_conv 

from core_types import IncompatibleTypes, ImmutableT

# base class for all concrete scalar types
# don't actually tag any values with this
class ScalarT(ImmutableT):
  rank = 0
  
  def __init__(self, dtype):
    self.dtype = dtype 
    self.name = self.dtype.name
    self.nbytes = self.dtype.itemsize
    self._hash = hash(self.dtype)
  
  def __eq__(self, other):
    return (self is other) or ( self.__class__ is other.__class__ and self.nbytes == other.nbytes)
  
  def __ne__(self, other):
    return not (self == other) 
  
  def __hash__(self):
    return self._hash

  def __repr__(self):
    return self.name

  def __str__(self):
    return str(self.name)

  def index_type(self, _):
    return self

  def convert_python_value(self, x):
    return self.dtype.type(x)
  
  def combine(self, other):
    if hasattr(other, 'dtype'):
      combined_dtype = np.promote_types(self.dtype, other.dtype)
      if combined_dtype == self.dtype: 
        return self
      elif combined_dtype == other.dtype:
        return other
      else:
        return from_dtype(combined_dtype)
    else:
      raise IncompatibleTypes(self, other)

_dtype_to_parakeet_type = {}
def register_scalar_type(ParakeetClass, dtype, equiv_python_types = []):
  parakeet_type = ParakeetClass(dtype)
  _dtype_to_parakeet_type[dtype] = parakeet_type

  python_types = [dtype.type] + equiv_python_types

  for python_type in python_types:
    type_conv.register(python_type, parakeet_type)

  return parakeet_type



class IntT(ScalarT):
  """Base class for bool, signed and unsigned"""
  pass 

  

class BoolT(IntT):
  """
  The type is called BoolT to distinguish it from its only instantiation called
  Bool.
  """
  
  def __init__(self, dtype):
    ScalarT.__init__(self, dtype)
    self.name = 'bool'

  def __eq__(self, other):
    return other.__class__ is BoolT

Bool = register_scalar_type(BoolT, dtypes.bool8, equiv_python_types = [bool])

class UnsignedT(IntT):
  pass 

UInt8 = register_scalar_type(UnsignedT, dtypes.uint8)

UInt16 = register_scalar_type(UnsignedT, dtypes.uint16)
UInt32 = register_scalar_type(UnsignedT, dtypes.uint32)
UInt64 = register_scalar_type(UnsignedT, dtypes.uint64)

class SignedT(IntT):
  pass 

Int8 = register_scalar_type(SignedT, dtypes.int8)
Int16 = register_scalar_type(SignedT, dtypes.int16) 
Int32 = register_scalar_type(SignedT, dtypes.int32)
Int64 = register_scalar_type(SignedT, dtypes.int64,
                             equiv_python_types = [int, long])

class Int24T(SignedT):
  """
  We don't actually support 24-bit integers, but they're useful  
  for implemented corner-case casting logic, such as summing booleans
  """
  def __init__(self):
    self.name = "int24"
    self.nbytes = 3
    self._hash = hash("int24")
    
  def combine(self, other):
    if isinstance(other, IntT):
      if other.nbytes <= self.nbytes:
        return Int32 
    return other 
  
  def __hash__(self):
    return self._hash 
  
  def __eq__(self, other):
    return self.__class__ is other.__class__ 
    
  def __ne__(self, other):
    return self.__class__ is not other.__class__
  
Int24 = Int24T()

class FloatT(ScalarT):
  pass 

Float32 = register_scalar_type(FloatT, dtypes.float32)
Float64 = register_scalar_type(FloatT, dtypes.float64,
                               equiv_python_types = [float])

def is_scalar_subtype(t1, t2):
  return isinstance(t1, ScalarT) and \
         isinstance(t2, ScalarT) and \
         ((t1 is t2) or 
          (t1 == t2) or 
          (t1.nbytes < t2.nbytes) or 
          (isinstance(t1, IntT) and isinstance(t2, FloatT)))

def from_dtype(dtype):
  if not isinstance(dtype, np.dtype):
    assert hasattr(dtype, 'dtype'), "Expected a dtype but %s" % dtype 
    value = dtype(0)
    dtype = value.dtype 
  parakeet_type = _dtype_to_parakeet_type.get(dtype)
  if parakeet_type is None:
    assert False, "Don't know how to make Parakeet scalar type from dtype %s" % dtype
  return parakeet_type

def from_char_code(c):
  numpy_type = np.typeDict[c]
  return from_dtype(np.dtype(numpy_type))

def is_scalar(t):
  return isinstance(t, ScalarT)

def all_scalars(ts):
  return all(is_scalar(t) for t in  ts)



import math 
import numpy as np
import parakeet
from parakeet.testing_helpers import run_local_tests, expect_each

xs = [0.1, 0.5, 2.0]

def test_sin():
  expect_each(parakeet.sin, np.sin, xs)
  expect_each(math.sin, math.sin, xs)
  expect_each(np.sin, np.sin, xs)


def test_parakeet_sinh():
  expect_each(parakeet.sinh, np.sinh, xs)
  expect_each(math.sinh, np.sinh, xs)
  expect_each(np.sinh, np.sinh, xs)


def test_cos():
  expect_each(parakeet.cos, np.cos, xs)
  expect_each(math.cos, np.cos, xs)
  expect_each(np.cos, np.cos, xs)

def test_cosh():
  expect_each(parakeet.cosh, np.cosh, xs)
  expect_each(math.cosh, np.cosh, xs)
  expect_each(np.cosh, np.cosh, xs)

def test_tan():
  expect_each(parakeet.tan, np.tan, xs)
  expect_each(math.tan, np.tan, xs)
  expect_each(np.tan, np.tan, xs)

def test_tanh():
  expect_each(parakeet.tanh, np.tanh, xs)
  expect_each(math.tanh, np.tanh, xs)
  expect_each(np.tanh, np.tanh, xs)
  
def test_log():
  expect_each(parakeet.log, np.log, xs)
  expect_each(math.log, np.log, xs)
  expect_each(np.log, np.log, xs)

def test_log10():
  expect_each(parakeet.log10, np.log10, xs)
  expect_each(math.log10, np.log10, xs)
  expect_each(np.log10, np.log10, xs)

def test_exp():
  expect_each(parakeet.exp, np.exp, xs)
  expect_each(math.exp, np.exp, xs)
  expect_each(np.exp, np.exp, xs)
  
if __name__ == '__main__':
  run_local_tests()



import numpy as np 

#
# Tensor rotation
# from Peter Mortensen's stackoverflow question 
# @ http://stackoverflow.com/questions/4962606/fast-tensor-rotation-with-numpy/18301915
#


from parakeet import jit 

@jit 
def rotT_loops(T, g):
    Tprime = np.zeros((3,3,3,3))
    for i in range(3):
        for j in range(3):
            for k in range(3):
                for l in range(3):
                    for ii in range(3):
                        for jj in range(3):
                            for kk in range(3):
                                for ll in range(3):
                                    gg = g[ii,i]*g[jj,j]*g[kk,k]*g[ll,l]
                                    Tprime[i,j,k,l] = Tprime[i,j,k,l] + gg*T[ii,jj,kk,ll]
    return Tprime

T = np.random.randn(3,3,3,3)
g = np.random.randn(3,3)

rotT_loops(T,g)




import numpy as np
from parakeet.testing_helpers import expect, run_local_tests

n = 8
vec_int = np.arange(n)
vec_bool = vec_int % 2
vec_float = vec_bool * 2.75
vectors = [vec_int, vec_bool, vec_float]

indices = np.arange(n)[vec_bool]

def idx(x,i):
  return x[i]

def test_1d_by_1d():
  for v in vectors:
    expect(idx, [v, indices], idx(v,indices))

mat_int = np.array([vec_int]*n)
mat_float = np.array([vec_float]*n)
mat_bool = np.array([vec_bool]*n)
matrices = [mat_int, mat_float, mat_bool]

def test_2d_by_idx():
  for m in matrices:
    expect(idx, [m, indices], idx(m, indices))

if __name__ == '__main__':
    run_local_tests()



from ..  import names, syntax
from ..analysis.use_analysis import use_count 
from ..ndtypes import ArrayT 
from ..transforms import inline, Transform 
from .. syntax import Var, Const,  Return, TypedFn, DataAdverb, Adverb
from .. syntax import IndexMap, IndexReduce, Map, Reduce, OuterMap 
from ..syntax.helpers import zero_i64, none 

def fuse(prev_fn, prev_fixed_args, next_fn, next_fixed_args, fusion_args):
  if syntax.helpers.is_identity_fn(next_fn):
    assert len(next_fixed_args) == 0
    return prev_fn, prev_fixed_args

  """
  Expects the prev_fn's returned value to be one or more of the arguments to
  next_fn. Any element in 'const_args' which is None gets replaced by the
  returned Var
  """
  fused_formals = []
  fused_input_types = []
  fused_type_env = prev_fn.type_env.copy()
  next_name = names.original(next_fn.name)
  prev_name = names.original(prev_fn.name)
  if prev_name.startswith("fused_") or next_name.startswith("fused_"):
    fused_name = names.fresh(prev_name + "_" + next_name)
  else:
    fused_name = names.fresh("fused_" + prev_name +  "_" + next_name)

  prev_closure_formals = prev_fn.arg_names[:len(prev_fixed_args)]

  for prev_closure_arg_name in prev_closure_formals:
    t = prev_fn.type_env[prev_closure_arg_name]
    fused_formals.append(prev_closure_arg_name)
    fused_input_types.append(t)

  next_closure_formals = next_fn.arg_names[:len(next_fixed_args)]
  for next_closure_arg_name in next_closure_formals:
    t = next_fn.type_env[next_closure_arg_name]
    new_name = names.refresh(next_closure_arg_name)
    fused_type_env[new_name] = t
    fused_formals.append(new_name)
    fused_input_types.append(t)

  prev_direct_formals = prev_fn.arg_names[len(prev_fixed_args):]
  for arg_name in prev_direct_formals:
    t = prev_fn.type_env[arg_name]
    fused_formals.append(arg_name)
    fused_input_types.append(t)

  prev_return_var, fused_body = \
      inline.replace_return_with_var(prev_fn.body,
                                     fused_type_env,
                                     prev_fn.return_type)
  # for now we're restricting both functions to have a single return at the
  # outermost scope
  inline_args = list(next_closure_formals)
  for arg in fusion_args:
    if arg is None:
      inline_args.append(prev_return_var)
    elif isinstance(arg, int):
      # positional arg which is not being fused out
      inner_name = next_fn.arg_names[arg]
      inner_type = next_fn.type_env[inner_name]
      new_name = names.refresh(inner_name)
      fused_formals.append(new_name)
      fused_type_env[new_name] = inner_type
      fused_input_types.append(inner_type)
      var = Var(new_name, inner_type)
      inline_args.append(var)
    else:
      assert arg.__class__ is Const, \
         "Only scalars can be spliced as literals into a fused fn: %s" % arg
      inline_args.append(arg)
  next_return_var = inline.do_inline(next_fn, inline_args,
                                     fused_type_env, fused_body)
  fused_body.append(Return(next_return_var))

  # we're not renaming variables that originate from the predecessor function
  new_fn = TypedFn(name = fused_name,
                   arg_names = fused_formals,
                   body = fused_body,
                   input_types = tuple(fused_input_types),
                   return_type = next_fn.return_type,
                   type_env = fused_type_env)

  combined_args = prev_fixed_args + next_fixed_args
  return new_fn, combined_args 

class Fusion(Transform):
  def __init__(self, recursive=True):
    Transform.__init__(self)
    # name of variable -> Map or Scan adverb
    self.adverb_bindings = {}
    self.recursive = True

  def pre_apply(self, fn):
    # map each variable to
    self.use_counts = use_count(fn)

  def transform_TypedFn(self, fn):
    return run_fusion(fn)
    #import pipeline
    #return pipeline.high_level_optimizations(fn)

  def has_required_rank(self, var_name, required_rank):
    if var_name not in self.type_env:
      return False 
    
    t = self.type_env[var_name]
     
    if t.__class__ is ArrayT:
      return t.rank == required_rank
    else:
      return required_rank == 0 
  
  def fuse_expr(self, rhs):
    if not isinstance(rhs, DataAdverb): return rhs 
    
    # TODO: figure out how to fuse OuterMap(Map(...), some_other_arg)
    # if rhs.__class__ is OuterMap: return stmt
     
    rhs_fn = self.get_fn(rhs.fn)

    if not inline.can_inline(rhs_fn): return rhs

    args = rhs.args
    if any(arg.__class__ not in (Var, Const) for arg in args): return rhs 
    
    arg_names = [arg.name for arg in args if arg.__class__ is Var]
    unique_vars = set(arg_names)
    
     
    valid_fusion_vars = [name for name in unique_vars
                         if name in self.adverb_bindings]

    # only fuse over variables which of the same rank as whatever is the largest
    # array being processed by this operator 
    max_rank = max(self.rank(arg) for arg in args)

    valid_fusion_vars = [name for name in valid_fusion_vars
                         if self.has_required_rank(name,max_rank)]
    
    for arg_name in valid_fusion_vars:
      n_occurrences = sum((name == arg_name for name in arg_names))
      prev_adverb = self.adverb_bindings[arg_name]
      prev_adverb_fn = self.get_fn(prev_adverb.fn)
      
      if not inline.can_inline(prev_adverb_fn):
        continue
      
      # if we're doing a cartesian product between inputs then 
      # can't introduce multiple new array arguments 
      if rhs.__class__ is OuterMap and len(prev_adverb.args) != 1:
        continue 
       
      # 
      # Map(Map) -> Map
      # Reduce(Map) -> Reduce 
      # Scan(Map) -> Scan
      # ...but for some reason, not:
      # OuterMap(Map) -> OuterMap 
      # 
      if prev_adverb.__class__ is Map and rhs.axis == prev_adverb.axis:
   
        surviving_array_args = []
        fusion_args = []
        for (pos, arg) in enumerate(rhs.args):
          c = arg.__class__
          if c is Var and arg.name == arg_name:
            fusion_args.append(None)
          elif c is Const:
            fusion_args.append(arg)
          else:
            surviving_array_args.append(arg)
            fusion_args.append(pos)
        
        new_fn, clos_args = \
              fuse(self.get_fn(prev_adverb.fn),
                   self.closure_elts(prev_adverb.fn),
                   self.get_fn(rhs.fn),
                   self.closure_elts(rhs.fn),
                   fusion_args)

        assert new_fn.return_type == self.return_type(rhs.fn)
        
        if self.use_counts[arg_name] == n_occurrences:
          del self.adverb_bindings[arg_name]
        if self.fn.created_by is not None:
          new_fn = self.fn.created_by.apply(new_fn)
        rhs.fn = self.closure(new_fn, clos_args)
        rhs.args = prev_adverb.args + surviving_array_args
      
      # 
      # Reduce(IndexMap) -> IndexReduce
      #   
      elif prev_adverb.__class__ is IndexMap and \
            rhs.__class__ is Reduce and \
            len(rhs.args) == 1 and \
            (self.is_none(rhs.axis) or rhs.args[0].type.rank == 1):
              
        new_fn, clos_args = \
            fuse(self.get_fn(prev_adverb.fn),
                 self.closure_elts(prev_adverb.fn),
                 self.get_fn(rhs.fn),
                 self.closure_elts(rhs.fn),
                 rhs.args)
        assert new_fn.return_type == self.return_type(rhs.fn)
        if self.use_counts[arg_name] == n_occurrences:
          del self.adverb_bindings[arg_name]
        
        if self.fn.created_by is not None:
          new_fn = self.fn.created_by.apply(new_fn)
        rhs = IndexReduce(fn = self.closure(new_fn, clos_args), 
                               shape = prev_adverb.shape, 
                               combine = rhs.combine, 
                               type = rhs.type,
                               init = rhs.init)
    return rhs 

    
  
  def transform_Assign(self, stmt):
    old_rhs = stmt.rhs 
    if self.recursive: 
      old_rhs = self.transform_expr(old_rhs)
    
    if not isinstance(old_rhs, DataAdverb):
      return stmt 

    new_rhs = self.fuse_expr(old_rhs)
    if stmt.lhs.__class__ is Var and isinstance(new_rhs, Adverb):
      self.adverb_bindings[stmt.lhs.name] = new_rhs
    stmt.rhs = new_rhs
    return stmt
  
  def transform_Return(self, stmt):
    old_rhs = stmt.value 
    if not isinstance(old_rhs, DataAdverb):
      return stmt 
    new_rhs = self.fuse_expr(old_rhs)
    stmt.value = new_rhs 
    return stmt 

  
from ..analysis import contains_adverbs, contains_calls
def run_fusion(fn):
  if contains_adverbs(fn) or contains_calls(fn):
    return Fusion().apply(fn)
  else:
    return fn 
      



import numpy as np

from parakeet import jit, allpairs 
from parakeet.testing_helpers import run_local_tests, expect

int_mat = np.reshape(np.arange(6), (2,3))
float_mat = np.sqrt(int_mat)
bool_mat = int_mat % 2

matrices = [int_mat, float_mat, bool_mat]
      
@jit 
def dot(x,y):
  return sum(x*y)

@jit 
def matmult_allpairs(X,Y):
  return allpairs(dot, X, Y, axis = (0,1))

def test_matmult_allpairs():
  for X in matrices:
    for Y in matrices:
      res = np.dot(X, Y.T)
      expect(matmult_allpairs, [X,Y.T], res)

      
if __name__ == "__main__":
  run_local_tests()



"""
SPH renderer, modified from Numba version at:
https://gist.github.com/rokroskar/bdcf6c6b210ff0efc738#file-gistfile1-txt-L55
"""
 
 
import numpy as np
from numpy import int32
from parakeet import testing_helpers 



 
def kernel_func(d, h) : 
    if d < 1 : 
        f = 1.-(3./2)*d**2 + (3./4.)*d**3
    elif d<2 :
        f = 0.25*(2.-d)**3
    else :
        f = 0
    return f/(np.pi*h**3)
 
def distance(x,y,z) : 
    return np.sqrt(x**2+y**2+z**2)
 
def physical_to_pixel(xpos,xmin,dx) : 
    return int32((xpos-xmin)/dx)
 
def pixel_to_physical(xpix,x_start,dx) : 
    return dx*xpix+x_start
 
def render_image(xs, ys, zs, hs, qts, mass, rhos, nx, ny, 
                 xmin = 0.0, xmax = 1.0, ymin = 0.0, ymax = 1.0): 
    MAX_D_OVER_H = 2.0
 
    image = np.zeros((nx,ny))
 
    dx = (xmax-xmin)/nx
    dy = (ymax-ymin)/ny
 
    x_start = xmin+dx/2
    y_start = ymin+dy/2
    zplane = 0.0
 
    # set up the kernel values
    kernel_samples = np.arange(0, 2.01, .01)
    kernel_vals = np.array([kernel_func(x,1.0) for x in kernel_samples])
    for i, (x,y,z,h) in enumerate(zip(xs,ys,zs,hs)):
        qt = qts[i] * mass[i] / rhos[i]

        # is the particle in the frame?
        if ((x > xmin-2*h) and (x < xmax+2*h) and 
            (y > ymin-2*h) and (y < ymax+2*h) and 
            (np.abs(z-zplane) < 2*h)) : 
        
                    
            if (MAX_D_OVER_H*h/dx < 1 ) and (MAX_D_OVER_H*h/dy < 1) : 
                # pixel coordinates 
                xpos = physical_to_pixel(x,xmin,dx)
                ypos = physical_to_pixel(y,ymin,dy)
                # physical coordinates of pixel
                xpixel = pixel_to_physical(xpos,x_start,dx)
                ypixel = pixel_to_physical(ypos,y_start,dy)
                zpixel = zplane
 
                dxpix, dypix, dzpix = [x-xpixel,y-ypixel,z-zpixel]
                d = distance(dxpix,dypix,dzpix)
                if (xpos > 0) and (xpos < nx) and (ypos > 0) and (ypos < ny) and (d/h < 2) : 
                    kernel_val = kernel_vals[int(d/(.01*h))]/(h*h*h)
                    image[xpos,ypos] += qt*kernel_val
 
            else :
                # bottom left of pixels the particle will contribute to
                x_pix_start = int32(physical_to_pixel(x-MAX_D_OVER_H*h,xmin,dx))
                x_pix_stop  = int32(physical_to_pixel(x+MAX_D_OVER_H*h,xmin,dx))
                y_pix_start = int32(physical_to_pixel(y-MAX_D_OVER_H*h,ymin,dy))
                y_pix_stop  = int32(physical_to_pixel(y+MAX_D_OVER_H*h,ymin,dy))
            
                if(x_pix_start < 0):  x_pix_start = 0
                if(x_pix_stop  > nx): x_pix_stop  = int32(nx-1)
                if(y_pix_start < 0):  y_pix_start = 0
                if(y_pix_stop  > ny): y_pix_stop  = int32(ny-1)
    
                
                for xpix in range(x_pix_start, x_pix_stop) : 
                    for ypix in range(y_pix_start, y_pix_stop) : 
                        # physical coordinates of pixel
                        xpixel = pixel_to_physical(xpix,x_start,dx)
                        ypixel = pixel_to_physical(ypix,y_start,dy)
                        zpixel = zplane
 
                        dxpix, dypix, dzpix = [x-xpixel,y-ypixel,z-zpixel]
                        d = distance(dxpix,dypix,dzpix)
                        if (d/h < 2) : 
                            kernel_val = kernel_vals[int(d/(.01*h))]/(h*h*h)
                            image[xpix,ypix] += qt*kernel_val
    
 
    return image
 
 
def test_image_render() : 
    N = 4
    x = y = z = hs= qts = mass = rhos = np.random.rand(N)
    nx=ny=2
    args = (x,y,z,hs,qts,mass,rhos,nx,ny)
    testing_helpers.expect(render_image, args, render_image(*args))

if __name__ == '__main__':
  testing_helpers.run_local_tests()



from .. import  prims 
from .. syntax import Var, PrimCall, Const
from syntax_visitor import SyntaxVisitor

class OffsetAnalysis(SyntaxVisitor):
  """Determine static offset relationships between variables"""
  
  def __init__(self):
    # map from variable 'x' to list of variable offset pairs
    # [ ('y', 4), ('z', -1), ...] 
    self.known_offsets = {}
    
  def update(self, x, y, k):
    if x == y:
      assert k == 0, \
         "Impossible %s = %s + %d" % (x,y,k)
    
    if x in self.known_offsets:
      x_offsets = self.known_offsets[x]
    else:
      x_offsets = set([])
      
    x_offsets.add( (y,k) )

    if y in self.known_offsets:
      y_offsets = self.known_offsets[y]
    else:
      y_offsets = set([])
      
    for (z, k2) in y_offsets:
      x_offsets.add( (z, k2 + k) )
      
    self.known_offsets[x] = x_offsets
    self.known_offsets[y] = y_offsets
    
  def visit_merge(self, merge):
    for (k, (l,r)) in merge.iteritems():
      if l.__class__ is Var and \
         r.__class__ is Var and \
         l.name in self.known_offsets and \
         r.name in self.known_offsets:
        left = self.known_offsets[l.name]
        right = self.known_offsets[r.name]
        self.known_offsets[k] = left.intersection(right)  
  
  def visit_PrimCall(self, expr):
    if expr.prim is prims.add:
      x, y = expr.args
      if x.__class__ is Var and y.__class__ is Const:
        return (x.name, y.value)
      elif y.__class__ is Var and x.__class__ is Const:
        return (y.name, x.value)
    elif expr.prim is prims.subtract:
      x, y = expr.args
      if x.__class__ is Var and y.__class__ is Const:
        return (x.name, -y.value)
    return None
  
  def visit_Assign(self, stmt):
    if stmt.lhs.__class__ is Var:
      if stmt.rhs.__class__ is PrimCall:
        rhs = self.visit_PrimCall(stmt.rhs)
        if rhs is not None:
          x = stmt.lhs.name
          (y, offset) = rhs
          self.update(x, y, offset)
          self.update(x, y, offset)
          self.update(y, x, -offset)
      elif stmt.rhs.__class__ is Var:
        x = stmt.lhs.name 
        y = stmt.rhs.name
        self.update(x, y, 0)
        self.update(y, x, 0)
  
  def visit_fn(self, fn):
    SyntaxVisitor.visit_fn(self, fn)
    return self.known_offsets




import numpy as np 
from parakeet.testing_helpers import expect, run_local_tests
from parakeet import config  

config.print_specialized_function = False
config.print_loopy_function = False


vec = np.arange(16)
mat = vec.reshape(4,4)

def negative_stop_1d(x):
    return x[:-2]

def test_negative_stop_1d_vec():
    expect(negative_stop_1d, [vec], negative_stop_1d(vec))

def test_negative_stop_1d_mat():
    expect(negative_stop_1d, [mat], negative_stop_1d(mat))

def negative_stop_2d(x):
    return x[:-2, :-3]

def test_negative_stop_2d():
    expect(negative_stop_2d, [mat], negative_stop_2d(mat))

def negative_start_1d(x):
    return x[-2:]

def test_negative_start_1d_vec():
    expect(negative_start_1d, [vec], vec[-2:])

def test_negative_start_1d_mat():
    expect(negative_start_1d, [mat], mat[-2:])

def negative_start_2d(x):
    return x[-2:, -3:]

def test_negative_start_2d():
    expect(negative_start_2d, [mat], mat[-2:, -3:])

def negative_step_1d(x):
    return x[::-2]

def test_negative_step_1d_vec():
    expect(negative_step_1d, [vec], vec[::-2])

def test_negative_step_1d_mat():
    expect(negative_step_1d, [mat], mat[::-2])

def negative_step_2d(x):
    return x[::-2, ::-3]

def test_negative_step_2d():
    expect(negative_step_2d, [mat], mat[::-2, ::-3])

if __name__ == "__main__":
    run_local_tests()





from collect_vars import collect_var_names, collect_var_names_from_exprs
from escape_analysis import escape_analysis
from syntax_visitor import SyntaxVisitor


class ArrayWriteAnalysis(SyntaxVisitor):
  def __init__(self, fn, fresh_alloc_args = set([])):
    self.fn = fn 
    self.fresh_alloc_args = fresh_alloc_args
    
  def visit_fn(self, fn):
    escape_info = escape_analysis(fn, self.fresh_alloc_args)
    self.may_alias = escape_info.may_alias 
    SyntaxVisitor.visit_fn(self, fn)
    self.writes = set([])
    
  def visit_Assign(self, stmt):
    if stmt.lhs.__class__ is Tuple:
      for elt in stmt.lhs.elts:
        self.visit_Assign(elt)
    elif stmt.lhs.__class__ is Index:
      for name in collect_var_names(stmt.lhs.value):
        self.writes.append(name)
        for alias_name in self.may_alias[name]:
          self.writes.append(alias_name)
      
    
  def visit_Index(self):
   



import numpy as np
from parakeet import jit 
from parakeet.testing_helpers import expect, expect_each, run_local_tests

shape_1d = 40
ints_1d = np.arange(shape_1d)
floats_1d = np.arange(shape_1d, dtype='float')
bools_1d = ints_1d % 2

vecs = [ints_1d, floats_1d, bools_1d]

shape_2d = (4,10)
matrices = [np.reshape(vec, shape_2d) for vec in vecs]

def implicit_slice_first_axis(x,i):
  return x[i]

def test_implicit_slice_first_axis_matrices():
  for m in matrices:
    expect(implicit_slice_first_axis, [m,2], m[2])

def slice_first_axis(x,i):
  return x[i,:]

def test_slice_first_axis_matrices():
  for m in matrices:
    expect(implicit_slice_first_axis, [m, 2], m[2])

def slice_second_axis(x,i):
  return x[:,i]

def test_slice_second_axis_matrices():
  for m in matrices:
    expect(slice_second_axis, [m,2], m[:,2])

def assign_first_axis(x, i, j):
  x[i] = x[j]
  return x

def test_assign_first_axis():
  for m in matrices:
    m_expect = m.copy()
    m_input = m.copy()
    m_expect[1] = m_expect[2]
    expect(assign_first_axis, [m_input, 1, 2], m_expect)

def assign_second_axis(x, i, j):
  x[:, i] = x[:, j]
  return x

def test_assign_second_axis():
  for m in matrices:
    m_expect = m.copy()
    m_input = m.copy()
    m_expect[:,1] = m_expect[:,2]
    expect(assign_second_axis, [m_input, 1, 2], m_expect)

def assign_slices(x, (i,j,k,l), (a,b,c,d)):
  x[i:j, k:l] = x[a:b, c:d]
  return x

def test_assign_slices():
  for m in matrices:
    m_expect = m.copy()
    m_input = m.copy()
    (i,j,k,l) = (0,2,0,4)
    (a,b,c,d) = (1,3,5,9)
    m_expect[i:j, k:l] = m_expect[a:b, c:d]
    expect(assign_slices, [m_input, (i,j,k,l), (a,b,c,d)], m_expect)
    expect(assign_slices, [m_input, (i,j,k,l), (a,b,c,d)], m_expect)

def assign_four_rows(x, y):
  y[0:2,0:5] = x[0:2,0:5]
  y[0:2,5:10] = x[0:2,5:10]
  y[2:4,0:5] = x[2:4,0:5]
  y[2:4,5:10] = x[2:4,5:10]
  return y

def test_assign_four_rows():
  for m in matrices:
    m_expect = m.copy()
    m_input = m.copy()
    m_zeros = np.zeros_like(m_input)
    expect(assign_four_rows, [m_input, m_zeros], m_expect)

def copy_two_rows(x, y):
  # Fill y with the elements of x to try to get an identical copy
  i = 0
  while i < x.shape[1]:
    y[0][i] = x[0][i]
    y[1][i] = x[1][i]
    i = i + 1
  return y

def loop_slice(x, y, z):
  i = 0
  while i < z.shape[0]:
    i_next = i + 2
    z[i:i_next,:] = copy_two_rows(x[i:i_next,:], y)
    i = i_next
  return z

def test_loop_slices():
  m_input = np.arange(40, dtype=np.int64).reshape(4,10)
  m_zeros = np.zeros((2,10), dtype=np.int64)
  m_z = np.zeros_like(m_input)
  m_expect = m_input.copy()
  expect(loop_slice, [m_input, m_zeros, m_z], m_expect)

def lower_right_corner(X):
  m,n = X.shape
  return X[m/2:m, n/2:n]

def test_lower_right_corner():
  expect_each(jit(lower_right_corner), lower_right_corner, matrices)

def multiple_slices(x):
  y = x[2:4,:]
  return y[:,2]

def test_multiple_slices():
  x = np.random.randn(10,10)
  expect_each(jit(multiple_slices), multiple_slices, matrices)
  

if __name__ == '__main__':
  run_local_tests()



import numpy as np 
from parakeet.testing_helpers import expect, run_local_tests

# Simple convolution of 3x3 patches from a given array x
# by a 3x3 array of filter weights
 
def conv_3x3_trim(x, weights):
  return np.array([[(x[i-1:i+2, j-1:j+2]*weights).sum() 
                    for j in xrange(1, x.shape[1] -1)]
                    for i in xrange(1, x.shape[0] -1)])
 
x = np.random.randn(4,4)
w = np.random.randn(3,3)
"""
def test_conv_float32():
    x_f32 = x.astype('float32')
    w_f32 = w.astype('float32')
    expect(conv_3x3_trim, [x_f32, w_f32], conv_3x3_trim(x_f32,w_f32))

def test_conv_int16():
    x_i16 = x.astype('int16')
    w_i16 = w.astype('int16')
    expect(conv_3x3_trim, [x_i16, w_i16], conv_3x3_trim(x_i16,w_i16))

def test_conv_bool():
    xb = x > 0
    wb = w > 0
    expect(conv_3x3_trim, [xb, wb], conv_3x3_trim(xb,wb))
"""
def conv_3x3_trim_loops(image, weights):
  result = np.zeros_like(image[1:-1, 1:-1])
  m,n = image.shape 
  for i in xrange(1,m-1):
    for j in xrange(1,n-1):
      for ii in xrange(3): 
        for jj in xrange(3):
          result[i-1,j-1] += image[i-ii+1, j-jj+1] * weights[ii, jj] 
  return result

def test_conv_loops_float32():
    x_f32 = x.astype('float32')
    w_f32 = w.astype('float32')
   
    expect(conv_3x3_trim_loops, [x_f32, w_f32], conv_3x3_trim_loops(x_f32,w_f32))

def test_conv_loops_int16():
    x_i16 = x.astype('int16')
    w_i16 = w.astype('int16')
    expect(conv_3x3_trim_loops, [x_i16, w_i16], conv_3x3_trim_loops(x_i16,w_i16))

def test_conv_loops_bool():
    xb = x > 0
    wb = w > 0
    expect(conv_3x3_trim_loops, [xb, wb], conv_3x3_trim_loops(xb,wb))

if __name__ == "__main__":
    run_local_tests()



"""
Kernel from SPH renderer, modified from Numba version at:
https://gist.github.com/rokroskar/bdcf6c6b210ff0efc738#file-gistfile1-txt-L55
"""
 
 
import numpy as np
from parakeet import testing_helpers, config 


def kernel_func(d, h) : 
    if d < 1 : 
        f = 1.-(3./2)*d**2 + (3./4.)*d**3
    elif d<2 :
        f = 0.25*(2.-d)**3
    else :
        f = 0
    return f/(np.pi*h**3)
 
 
def compute_kernel( start = 0, stop = 2.01, step = 0.01,  h = 1.0): 
    # set up the kernel values
    kernel_samples = np.arange(start, stop, step)
    return np.array([kernel_func(x, h) for x in kernel_samples])

 
print compute_kernel(0, 0.1, 0.05)

def test_kernel():
  testing_helpers.expect(compute_kernel, [0, 0.5], compute_kernel(0,0.5))

if __name__ == "__main__":
  testing_helpers.run_local_tests()




from parakeet import jit 

@jit
def sum_loop(xs):
    total = 0
    for x in xs:
        total += x
    return total 

xs = [1,2,3,4,5]
print "Python result", sum(xs)
print "Parakeet result", sum_loop(xs)



from .. syntax import Expr, Var
from clone_function import CloneFunction
from transform import Transform

class RewriteVars(Transform):
  def __init__(self, rename_dict):
    Transform.__init__(self, require_types = False)
    self.rename_dict = rename_dict

  def transform_merge(self, old_merge):
    new_merge = {}
    for (k,(l,r)) in old_merge.iteritems():
      new_name = self.rename_dict.get(k,k)
      if type(new_name) != str:
        assert new_name.__class__ is Var, \
            "Unexpected substitution %s for %s" % (new_name, k)
        new_name = new_name.name
      new_left = self.transform_expr(l)
      new_right = self.transform_expr(r)
      new_merge[new_name] = new_left, new_right
    return new_merge
  
  def transform_Var(self, expr):
    new_value = self.rename_dict.get(expr.name, expr.name)
    if new_value.__class__ is str:
      if new_value != expr.name:
        expr.name = new_value
      return expr
    else:
      assert isinstance(expr, Expr)
      assert new_value.type is not None, \
          "Type of replacement value %s can't be None" % new_value
      return new_value

def subst_expr(expr, rename_dict):
  fresh_expr = CloneFunction().transform_expr(expr)
  return RewriteVars(rename_dict).transform_expr(fresh_expr)

def subst_expr_list(nodes, rename_dict):
  return [subst_expr(node, rename_dict) for node in nodes]

def subst_expr_tuple(elts, rename_dict):
  return tuple(subst_expr_list(elts, rename_dict))

def subst_stmt_list(stmts, rename_dict):
  fresh_stmts = CloneFunction().transform_block(stmts)
  return RewriteVars(rename_dict).transform_block(fresh_stmts)



import __builtin__
import ast
import inspect
import types
from collections import OrderedDict

import numpy as np
from dsltools import NestedBlocks, ScopedDict
 
from .. import config, names, prims, syntax

from ..names import NameNotFound
from ..ndtypes import Type
from ..prims import Prim 
from ..syntax import (Expr, 
                      Assign, If, ForLoop, Return,  
                      Var, PrimCall, Cast,  Select, 
                      Map, Reduce, 
                      Enumerate, Zip, Len, Range,    
                      Slice,  Tuple, Array,
                      Const, Call, Index, 
                      FormalArgs, ActualArgs, 
                      UntypedFn, Closure, 
                      SourceInfo)
 
from ..syntax.helpers import (none, true, false, one_i64, zero_i64, zero_i24,  
                              is_python_constant, const)
from ..syntax.wrappers import build_untyped_prim_fn, build_untyped_expr_fn, build_untyped_cast_fn


 
from ..transforms import subst_expr, subst_stmt_list

from decorators import jit, macro  
from python_ref import GlobalValueRef,  ClosureCellRef


class UnsupportedSyntax(Exception):
  def __init__(self, node, function_name = None, filename = None):
    self.function_name = function_name 
    self.filename = filename 
    self.node = node
    
  def __str__(self):
    if self.function_name is not None and \
       self.filename is not None and \
       self.node.lineno is not None:
      return "Parakeet doesn't support %s from function '%s' in file %s on line %d" % \
        (self.node.__class__.__name__, self.function_name, self.filename, self.node.lineno)
    elif self.function_name is not None:
      return "Parakeet doesn't support %s in '%s'" % \
        (self.node.__class__.__name__, self.function_name)
    else:
      return "Parakeet doesn't support %s" % self.node.__class__.__name__


  
class ExternalValue(object):
  """
  Wrap up references global values with this class
  """ 
  def __init__(self, python_value):
    self.value = python_value
    
  def __str__(self):
    return "ExternalValue(%s)" % self.value


def mk_reduce_call(fn, positional, init = None):
  init = none if init is None else init
  axis = zero_i64
  from .. import lib 
  return Reduce(fn = translate_function_value(lib.identity), 
                combine = fn, 
                args = positional, 
                axis = axis, 
                init = init)

def mk_simple_fn(mk_body, input_name = "x", fn_name = "cast"):
  unique_arg_name = names.fresh(input_name)
  unique_fn_name = names.fresh(fn_name)
  var = Var(unique_arg_name)
  formals = FormalArgs()
  formals.add_positional(unique_arg_name, input_name)
  body = mk_body(var)
  return UntypedFn(unique_fn_name, formals, body)
  
       
def is_hashable(x):
  try:
    hash(x)
    return True
  except:
    return False 

def is_prim(v):
  return isinstance(v, (Prim)) or (is_hashable(v) and v in prims.prim_lookup_by_value) 

def is_builtin_function(v):
  return isinstance(v, (types.TypeType, types.BuiltinFunctionType))
  
def is_user_function(v):
  return isinstance(v, (types.FunctionType, jit, macro))

def is_function_value(v):
  return is_user_function(v) or is_builtin_function(v) or is_prim(v) 
    
def is_static_value(v):
  return is_python_constant(v) or \
         type(v) is np.dtype  or \
         is_function_value(v)

def value_to_syntax(v):
  if isinstance(v, Expr):
    return v 
  elif is_python_constant(v):
    return const(v)
  else:
    return translate_function_value(v)
    
class AST_Translator(ast.NodeVisitor):
  def __init__(self, 
               globals_dict=None, 
               closure_cell_dict=None,
               parent=None, 
               function_name = None, 
               filename = None):
    # assignments which need to get prepended at the beginning of the
    # function
    self.globals = globals_dict
    self.blocks = NestedBlocks()

    self.parent = parent 
    self.scopes = ScopedDict()
    
    self.globals_dict = globals_dict 
    self.closure_cell_dict = closure_cell_dict 
    
    # mapping from names/paths to either a closure cell reference or a 
    # global value 
    self.python_refs = OrderedDict()
    
    self.original_outer_names = []
    self.localized_outer_names = []
    self.filename = filename 
    self.function_name = function_name 
    self.push()

  def push(self, scope = None, block = None):
    if scope is None:
      scope = {}
    if block is None:
      block = []
    self.scopes.push(scope)
    self.blocks.push(block)


  def pop(self):
    scope = self.scopes.pop()
    block = self.blocks.pop()
    return scope, block
  
  def fresh_name(self, original_name):
    fresh_name = names.fresh(original_name)
    self.scopes[original_name] = fresh_name
    return fresh_name
    
  def fresh_names(self, original_names):
    return map(self.fresh_name, original_names)

  def fresh_var(self, name):
    return Var(self.fresh_name(name))

  def fresh_vars(self, original_names):
    return map(self.fresh_var, original_names)
  
  def current_block(self):
    return self.blocks.top()

  def current_scope(self):
    return self.scopes.top()

  def ast_to_value(self, expr):
    if isinstance(expr, ast.Num):
      return expr.n
    elif isinstance(expr, ast.Tuple):
      return tuple(self.ast_to_value(elt) for elt in expr.elts)
    elif isinstance(expr, ast.Name):
      return self.lookup_global(expr.id) 
    elif isinstance(expr, ast.Attribute):
      left = self.ast_to_value(expr.value)
      if isinstance(left, ExternalValue):
        left = left.value 
      return getattr(left, expr.attr) 

  def lookup_global(self, key):
    if isinstance(key, (list, tuple)):
      assert len(key) == 1
      key = key[0]
    else:
      assert isinstance(key, str), "Invalid global key: %s" % (key,)

    if self.globals:
      if key in self.globals:
        return self.globals[key]
      elif key in __builtin__.__dict__:

        return __builtin__.__dict__[key]
      else:
        assert False, "Couldn't find global name %s" % key
    else:
      assert self.parent is not None
      return self.parent.lookup_global(key)
    
  def is_global(self, key):
    if isinstance(key, (list, tuple)):
      key = key[0]
    if key in self.scopes:
      return False
    elif self.closure_cell_dict and key in self.closure_cell_dict:
      return False
    if self.globals:
      return key in self.globals or key in __builtins__  
    assert self.parent is not None 
    return self.parent.is_global(key)
  
  
  def local_ref_name(self, ref, python_name):
    for (local_name, other_ref) in self.python_refs.iteritems():
      if ref == other_ref:
        return Var(local_name)
    local_name = names.fresh(python_name)
    self.scopes[python_name] = local_name
    self.original_outer_names.append(python_name)
    self.localized_outer_names.append(local_name)
    self.python_refs[local_name] = ref
    return Var(local_name)
  
  def is_visible_name(self, name):
    if name in self.scopes:
      return True 
    if self.parent:
      return self.parent.is_visible_name(name)
    else:
      return self.is_global(name)
      
  def lookup(self, name):
    #if name in reserved_names:
    #  return reserved_names[name]
    if name in self.scopes:
      return Var(self.scopes[name])
    elif self.parent and self.parent.is_visible_name(name):
      # don't actually keep the outer binding name, we just
      # need to check that it's possible and tell the outer scope
      # to register any necessary python refs
      local_name = names.fresh(name)
      self.scopes[name] = local_name
      self.original_outer_names.append(name)
      self.localized_outer_names.append(local_name)
      return Var(local_name)
    elif self.closure_cell_dict and name in self.closure_cell_dict:
      ref = ClosureCellRef(self.closure_cell_dict[name], name)
      return self.local_ref_name(ref, name)
    elif self.is_global(name):
      value = self.lookup_global(name)
      if is_static_value(value):
        return value_to_syntax(value)
      elif isinstance(value, np.ndarray): 
        ref = GlobalValueRef(value)
        return self.local_ref_name(ref, name)
      else:
        # assume that this is a module or object which will have some 
        # statically convertible value pulled out of it 
        return ExternalValue(value)
      #else:
      #  assert False, "Can't use global value %s" % value
    else:
      raise NameNotFound(name)
      
      
  def visit_list(self, nodes):
    return map(self.visit, nodes)


  def tuple_arg_assignments(self, elts, var):
    """
    Recursively decompose a nested tuple argument like
      def f((x,(y,z))):
        ...
    into a single name and a series of assignments:
      def f(tuple_arg):
        x = tuple_arg[0]
        tuple_arg_elt = tuple_arg[1]
        y = tuple_arg_elt[0]
        z = tuple_arg_elt[1]
    """

    assignments = []
    for (i, sub_arg) in enumerate(elts):
      if isinstance(sub_arg, ast.Tuple):
        name = "tuple_arg_elt"
      else:
        assert isinstance(sub_arg, ast.Name)
        name = sub_arg.id
      lhs = self.fresh_var(name)
      stmt = Assign(lhs, Index(var, Const(i)))
      assignments.append(stmt)
      if isinstance(sub_arg, ast.Tuple):
        more_stmts = self.tuple_arg_assignments(sub_arg.elts, lhs)
        assignments.extend(more_stmts)
    return assignments

  def translate_args(self, args):
    assert not args.kwarg
    formals = FormalArgs()
    assignments = []
    for arg in args.args:
      if isinstance(arg, ast.Name):
        visible_name = arg.id
        local_name = self.fresh_name(visible_name)
        formals.add_positional(local_name, visible_name)
      else:
        assert isinstance(arg, ast.Tuple)
        arg_name = self.fresh_name("tuple_arg")
        formals.add_positional(arg_name)
        var = Var(arg_name)
        stmts = self.tuple_arg_assignments(arg.elts, var)
        assignments.extend(stmts)

    n_defaults = len(args.defaults)
    if n_defaults > 0:
      local_names = formals.positional[-n_defaults:]
      for (k,expr) in zip(local_names, args.defaults):
        v = self.ast_to_value(expr)
        
        # for now we're putting literal python 
        # values in the defaults dictionary of
        # a function's formal arguments
        formals.defaults[k] = v 

    if args.vararg:
      assert isinstance(args.vararg, str)
      formals.starargs = self.fresh_name(args.vararg)

    return formals, assignments


  def visit_Name(self, expr):
    assert isinstance(expr, ast.Name), "Expected AST Name object: %s" % expr
    return self.lookup(expr.id)

  def create_phi_nodes(self, left_scope, right_scope, new_names = {}):
    """
    Phi nodes make explicit the possible sources of each variable's values and
    are needed when either two branches merge or when one was optionally taken.
    """
    merge = {}
    for (name, ssa_name) in left_scope.iteritems():
      left = Var(ssa_name)
      if name in right_scope:
        right = Var(right_scope[name])
      else:
        try:
          right = self.lookup(name)
        except NameNotFound:
          continue 
          
      if name in new_names:
        new_name = new_names[name]
      else:
        new_name = self.fresh_name(name)
      merge[new_name] = (left, right)

    for (name, ssa_name) in right_scope.iteritems():
      if name not in left_scope:
        try:
          left = self.lookup(name)
          right = Var(ssa_name)

          if name in new_names:
            new_name = new_names[name]
          else:
            new_name = self.fresh_name(name)
          merge[new_name] = (left, right)
        except names.NameNotFound:
          # for now skip over variables which weren't defined before
          # a control flow split, which means that loop-local variables
          # can't be used after the loop.
          # TODO: Fix this. Maybe with 'undef' nodes?
          pass
    return merge


  def visit_Index(self, expr):
    return self.visit(expr.value)

  def visit_Ellipsis(self, expr):
    raise RuntimeError("Ellipsis operator unsupported")

  def visit_Slice(self, expr):
    """
    x[l:u:s]
    Optional fields
      expr.lower
      expr.upper
      expr.step
    """

    start = self.visit(expr.lower) if expr.lower else none
    stop = self.visit(expr.upper) if expr.upper else none
    step = self.visit(expr.step) if expr.step else none
    return Slice(start, stop, step)

  def visit_ExtSlice(self, expr):
    slice_elts = map(self.visit, expr.dims)
    if len(slice_elts) > 1:
      return Tuple(slice_elts)
    else:
      return slice_elts[0]

  def visit_UnaryOp(self, expr):
    ssa_val = self.visit(expr.operand)
    # UAdd doesn't do anything!
    
    if expr.op.__class__.__name__ == 'UAdd':
      return ssa_val 
    prim = prims.find_ast_op(expr.op)
    return PrimCall(prim, [ssa_val])

  def visit_BinOp(self, expr):
    ssa_left = self.visit(expr.left)
    ssa_right = self.visit(expr.right)
    prim = prims.find_ast_op(expr.op)
    return PrimCall(prim, [ssa_left, ssa_right])

  def visit_BoolOp(self, expr):
    values = map(self.visit, expr.values)
    prim = prims.find_ast_op(expr.op)
    # Python, strangely, allows more than two arguments to
    # Boolean operators
    result = values[0]
    for v in values[1:]:
      result = PrimCall(prim, [result, v])
    return result

  def visit_Compare(self, expr):
    lhs = self.visit(expr.left)
    assert len(expr.ops) == 1
    prim = prims.find_ast_op(expr.ops[0])
    assert len(expr.comparators) == 1
    rhs = self.visit(expr.comparators[0])
    return PrimCall(prim, [lhs, rhs])

  def visit_Subscript(self, expr):
    value = self.visit(expr.value)
    index = self.visit(expr.slice)
    return Index(value, index)

  def generic_visit(self, expr):
    raise UnsupportedSyntax(expr, 
                            function_name = self.function_name, 
                            filename = self.filename)
  
  def visit(self, node):
    res = ast.NodeVisitor.visit(self, node)
    source_info = SourceInfo(filename = self.filename, 
                             line = getattr(node, 'lineno', None),
                             col = getattr(node, 'e.col_offset', None), 
                             function = self.function_name, )
    res.source_info = source_info 
    return res 
    
  def translate_value_call(self, value, positional, keywords_dict= {}, starargs_expr = None):
    if value is sum:
      return mk_reduce_call(build_untyped_prim_fn(prims.add), positional, zero_i24)
    
    elif value is max:
      if len(positional) == 1:
        return mk_reduce_call(build_untyped_prim_fn(prims.maximum), positional)
      else:
        assert len(positional) == 2
        return PrimCall(prims.maximum, positional)
    
    elif value is min:
      if len(positional) == 1:
        return mk_reduce_call(build_untyped_prim_fn(prims.minimum), positional)
      else:
        assert len(positional) == 2
        return PrimCall(prims.minimum, positional)
    
    elif value is map:
      assert len(keywords_dict) == 0
      assert len(positional) > 1
      axis = keywords_dict.get("axis", None)
      return Map(fn = positional[0], args = positional[1:], axis = axis)
    
    elif value is enumerate:
      assert len(positional) == 1, "Wrong number of args for 'enumerate': %s" % positional 
      assert len(keywords_dict) == 0, \
        "Didn't expect keyword arguments for 'enumerate': %s" % keywords_dict
      return Enumerate(positional[0])
    
    elif value is len:
      assert len(positional) == 1, "Wrong number of args for 'len': %s" % positional 
      assert len(keywords_dict) == 0, \
        "Didn't expect keyword arguments for 'len': %s" % keywords_dict
      return self.len(positional[0])
    
    elif value is zip:
      assert len(positional) > 1, "Wrong number of args for 'zip': %s" % positional 
      assert len(keywords_dict) == 0, \
        "Didn't expect keyword arguments for 'zip': %s" % keywords_dict
      return Zip(values = positional)
    
    from ..mappings import function_mappings
    if value in function_mappings:
      value = function_mappings[value]
    
    if isinstance(value, macro):
      return value.transform(positional, keywords_dict)
      
    fn = translate_function_value(value)
    return Call(fn, ActualArgs(positional, keywords_dict, starargs_expr))
    
  def visit_Call(self, expr):
    """
    TODO: 
    The logic here is broken and haphazard, eventually try to handle nested
    scopes correctly, along with globals, cell refs, etc..
    """
    
    fn, args, keywords_list, starargs, kwargs = \
        expr.func, expr.args, expr.keywords, expr.starargs, expr.kwargs
    assert kwargs is None, "Dictionary of keyword args not supported"

    positional = self.visit_list(args)

    keywords_dict = {}
    for kwd in keywords_list:
      keywords_dict[kwd.arg] = self.visit(kwd.value)

    if starargs:
      starargs_expr = self.visit(starargs)
    else:
      starargs_expr = None
    
    def is_attr_chain(expr):
      return isinstance(expr, ast.Name) or \
             (isinstance(expr, ast.Attribute) and is_attr_chain(expr.value))
    def extract_attr_chain(expr):
      if isinstance(expr, ast.Name):
        return [expr.id]
      else:
        base = extract_attr_chain(expr.value)
        base.append(expr.attr)
        return base
    
    def lookup_attr_chain(names):
      value = self.lookup_global(names[0])
    
      for name in names[1:]:
        if hasattr(value, name):
          value = getattr(value, name)
        else:
          try:
            value = value[name]
          except:
            assert False, "Couldn't find global name %s" % ('.'.join(names))
      return value
          
    if is_attr_chain(fn):
      names = extract_attr_chain(fn)
      
      if self.is_global(names):
        return self.translate_value_call(lookup_attr_chain(names), 
                                         positional, keywords_dict, starargs_expr)
    fn_node = self.visit(fn)    
    if isinstance(fn_node, syntax.Expr):
      actuals = ActualArgs(positional, keywords_dict, starargs_expr)
      return Call(fn_node, actuals)
    else:
      assert isinstance(fn_node, ExternalValue)
      return self.translate_value_call(fn_node.value, 
                                       positional, keywords_dict, starargs_expr)

  def visit_List(self, expr):
    return Array(self.visit_list(expr.elts))
    
  def visit_Expr(self, expr):
    # dummy assignment to allow for side effects on RHS
    lhs = self.fresh_var("dummy")
    if isinstance(expr.value, ast.Str):
      return Assign(lhs, zero_i64)
      # return syntax.Comment(expr.value.s.strip().replace('\n', ''))
    else:
      rhs = self.visit(expr.value)
      return syntax.Assign(lhs, rhs)

  def visit_GeneratorExp(self, expr):
    return self.visit_ListComp(expr)
    
  def visit_ListComp(self, expr):
    gens = expr.generators
    assert len(gens) == 1
    gen = gens[0]
    target = gen.target
    if target.__class__ is ast.Name:
      arg_vars = [target]
    else:
      assert target.__class__ is ast.Tuple and all(e.__class__ is ast.Name for e in target.elts),\
       "Expected comprehension target to be variable or tuple of variables, got %s" % ast.dump(target)
      arg_vars = [ast.Tuple(elts = target.elts)]
    # build a lambda as a Python ast representing 
    # what we do to each element 

    args = ast.arguments(args = arg_vars, 
                         vararg = None,  
                         kwarg = None,  
                         defaults = ())

    fn = translate_function_ast(name = "comprehension_map", 
                                args = args, 
                                body = [ast.Return(expr.elt)], 
                                parent = self)
    seq = self.visit(gen.iter)
    ifs = gen.ifs
    assert len(ifs) == 0, "Parakeet: Conditions in array comprehensions not yet supported"
    return Map(fn = fn, args=(seq,), axis = zero_i64)
      
  def visit_Attribute(self, expr):
    # TODO:
    # Recursive lookup to see if:
    #  (1) base object is local, if so-- create chain of attributes
    #  (2) base object is global but an adverb primitive-- use it locally
    #      without adding it to nonlocals
    #  (3) not local at all-- in which case, add the whole chain of strings
    #      to nonlocals
    #
    #  AN IDEA:
    #     Allow external values to be brought into the syntax tree as 
    #     a designated ExternalValue node
    #     and then here check if the LHS is an ExternalValue and if so, 
    #     pull out the value. If it's a constant, then make it into syntax, 
    #     if it's a function, then parse it, else raise an error. 
    #
    
    from ..mappings import property_mappings, method_mappings
    value = self.visit(expr.value)
    attr = expr.attr
    if isinstance(value, ExternalValue):
      value = value.value 
      assert hasattr(value, attr), "Couldn't find attribute '%s' in %s" % (attr, value)
      value = getattr(value, attr)
      if is_static_value(value):
        return value_to_syntax(value)
      else:
        return ExternalValue(value) 
    elif attr in property_mappings:
      fn = property_mappings[attr]
      if isinstance(fn, macro):
        return fn.transform( [value] )
      else:
        return Call(translate_function_value(fn),
                    ActualArgs(positional = (value,)))  
    elif attr in method_mappings:
      fn_python = method_mappings[attr]
      fn_syntax = translate_function_value(fn_python)
      return Closure(fn_syntax, args=(value,))
    else:
      assert False, "Attribute %s not supported" % attr 


  def visit_Num(self, expr):
    return Const(expr.n)

  def visit_Tuple(self, expr):
    return syntax.Tuple(self.visit_list(expr.elts))

  def visit_IfExp(self, expr):
    cond = self.visit(expr.test)
    if_true = self.visit(expr.body)
    if_false = self.visit(expr.orelse)
    return Select(cond, if_true, if_false)
    
  def visit_lhs(self, lhs):
    
    if isinstance(lhs, ast.Name):
      return  self.fresh_var(lhs.id)
    elif isinstance(lhs, ast.Tuple):
      return syntax.Tuple( map(self.visit_lhs, lhs.elts))
    else:
      # in case of slicing or attributes
      res = self.visit(lhs)
      return res

  def visit_Assign(self, stmt):  
    # important to evaluate RHS before LHS for statements like 'x = x + 1'
    ssa_rhs = self.visit(stmt.value)
    ssa_lhs = self.visit_lhs(stmt.targets[0])
    return Assign(ssa_lhs, ssa_rhs)
  
  def visit_AugAssign(self, stmt):
    ssa_incr = self.visit(stmt.value)
    ssa_old_value = self.visit(stmt.target)
    ssa_new_value = self.visit_lhs(stmt.target)
    prim = prims.find_ast_op(stmt.op) 
    return Assign(ssa_new_value, PrimCall(prim, [ssa_old_value, ssa_incr]))

  def visit_Return(self, stmt):
    return syntax.Return(self.visit(stmt.value))

  def visit_If(self, stmt):
    cond = self.visit(stmt.test)
    true_scope, true_block  = self.visit_block(stmt.body)
    false_scope, false_block = self.visit_block(stmt.orelse)
    merge = self.create_phi_nodes(true_scope, false_scope)
    return syntax.If(cond, true_block, false_block, merge)

  def visit_loop_body(self, body, *exprs):
    merge = {}
    substitutions = {}
    curr_scope = self.current_scope()
    exprs = [self.visit(expr) for expr in exprs]
    scope_after, body = self.visit_block(body)
    for (k, name_after) in scope_after.iteritems():
      if k in self.scopes:
        name_before = self.scopes[k]
        new_name = names.fresh(k + "_loop")
        merge[new_name] = (Var(name_before), Var(name_after))
        substitutions[name_before]  = new_name
        curr_scope[k] = new_name
    
    exprs = [subst_expr(expr, substitutions) for expr in exprs]
    body = subst_stmt_list(body, substitutions)
    return body, merge, exprs 

  def visit_While(self, stmt):
    assert not stmt.orelse
    body, merge, (cond,) = self.visit_loop_body(stmt.body, stmt.test)
    return syntax.While(cond, body, merge)

  def assign(self, lhs, rhs):
    self.current_block().append(Assign(lhs,rhs))
 
  def assign_to_var(self, rhs, name = None):
    if isinstance(rhs, (Var, Const)):
      return rhs 
    if name is None:
      name = "temp"
    var = self.fresh_var(name)
    self.assign(var, rhs)
    return var 
  
    
  def add(self, x, y, temp = True):
    expr = PrimCall(prims.add, [x,y])
    if temp:
      return self.assign_to_var(expr, "add")
    else:
      return expr 
  
  def sub(self, x, y, temp = True):
    expr = PrimCall(prims.subtract, [x,y])
    if temp:
      return self.assign_to_var(expr, "sub")
    else:
      return expr 
  
  def mul(self, x, y, temp = True):
    expr = PrimCall(prims.multiply, [x,y])
    if temp:
      return self.assign_to_var(expr, "mul")
    else:
      return expr 
  
  def div(self, x, y, temp = True):
    expr = PrimCall(prims.divide, [x,y])
    if temp:
      return self.assign_to_var(expr, "div")
    else:
      return expr 
  
  
  def len(self, x):
    if isinstance(x, Enumerate):
      return self.len(x.value)
    elif isinstance(x, Zip):
      elt_lens = [self.len(v) for v in x.values]
      result = elt_lens[0]
      for n in elt_lens[1:]:
        result = PrimCall(prims.minimum, [result, n])
      return result 
    elif isinstance(x, (Array, Tuple)):
      return Const(len(x.elts))
    
    elif isinstance(x, Range):
      # if it's a range from 0..len(x), then just return len(x)
      if isinstance(x.stop, Len):
        if isinstance(x.start, Const) and x.start.value == 0:
          if isinstance(x.step, Const) and x.stop.value in (1,-1, None):
            return x.stop
    seq_var = self.assign_to_var(x, "len_input")
    return self.assign_to_var(Len(seq_var), "len_result")


  def is_none(self, v):
    return v is None or isinstance(v, Const) and v.value is None 
  
  def for_loop_bindings(self, idx, lhs, rhs):
    if isinstance(rhs, Enumerate):
      array = rhs.value 
      elt = Index(array, idx)
      if isinstance(lhs, Tuple):
        var_names = ", ".join(str(elt) for elt in lhs.elts)
        if len(lhs.elts) < 2:
          raise SyntaxError("Too many values to unpack: 'enumerate' expects 2 but given %s" % var_names)
        elif len(lhs.elts) > 2:
          raise SyntaxError("Need more than 2 values to unpack for LHS of %s" % var_names)
        idx_var, seq_var = lhs.elts
        other_bindings = self.for_loop_bindings(idx, seq_var, array)
        return [Assign(idx_var, idx)] +  other_bindings 
      elif isinstance(lhs, Var):
        seq_var = self.fresh_var("seq_elt")
        other_bindings = self.for_loop_bindings(idx, seq_var, array)
        return [Assign(lhs, Tuple(idx, seq_var))] + other_bindings
      else:
        raise SyntaxError("Unexpected binding in for loop: %s = %s" % (lhs,rhs)) 
      
    elif isinstance(rhs, Zip):
      values_str = ", ".join(str(v) for v in rhs.values)
      if len(rhs.values) < 2:
        raise SyntaxError("'zip' must take at least two arguments, given: %s" % values_str)
      if isinstance(lhs, Tuple):
        if len(lhs.elts) < len(rhs.values):
          raise SyntaxError("Too many values to unpack in %s = %s" % (lhs, rhs))
        elif len(lhs.elts) > len(rhs.values):
          raise SyntaxError("Too few values on LHS of bindings in %s = %s" % (lhs,rhs))
        result = []
        for lhs_var, rhs_value in zip(lhs.elts, rhs.values):
          result.extend(self.for_loop_bindings(idx, lhs_var, rhs_value))
        return result 
      elif isinstance(lhs, Var):
        lhs_vars = [self.fresh_var("elt%d" % i) for i in xrange(len(rhs.values))]
        result = []
        for lhs_var, rhs_value in zip(lhs_vars, rhs.values):
          result.extend(self.for_loop_bindings(idx, lhs_var, rhs_value))
        result.append(Assign(lhs, Tuple(elts=lhs_vars)))
        return result  
      else:
        raise SyntaxError("Unexpected binding in for loop: %s = %s" % (lhs,rhs)) 
      
    elif isinstance(rhs, Range):
      if isinstance(lhs, Tuple):
        raise SyntaxError("Too few values in unpack in for loop binding %s = %s" % (lhs,rhs))
      elif isinstance(lhs, Var):
        start = rhs.start
        if self.is_none(start): 
          start = zero_i64
        step = rhs.step
        if self.is_none(step): 
          step = one_i64 
        return [Assign(lhs, self.add(start, self.mul(idx, step, temp = False), temp= False))]
      else:
        raise SyntaxError("Unexpected binding in for loop: %s = %s" % (lhs,rhs)) 
      
    else:
      return [Assign(lhs, Index(rhs,idx))]
      
      
  def visit_For(self, stmt):
    assert not stmt.orelse 
    var = self.visit_lhs(stmt.target)
    seq = self.visit(stmt.iter)
    body, merge, _ = self.visit_loop_body(stmt.body)

    if isinstance(seq, Range):
      assert isinstance(var, Var), "Expect loop variable to be simple but got '%s'" % var
      return ForLoop(var, seq.start, seq.stop, seq.step, body, merge)
    else:
      idx = self.fresh_var("idx")
      n = self.len(seq)
      bindings = self.for_loop_bindings(idx, var, seq)
      return ForLoop(idx, zero_i64, n, one_i64, bindings + body, merge)
    
  def visit_block(self, stmts):
    self.push()
    curr_block = self.current_block()
    for stmt in stmts:
      parakeet_stmt = self.visit(stmt)
      curr_block.append(parakeet_stmt)
    return self.pop()

  def visit_FunctionDef(self, node):
    """
    Translate a nested function
    """
    fundef = translate_function_ast(node.name, node.args, node.body, parent = self)
    local_var = self.fresh_var(node.name)
    return Assign(local_var, fundef)

  def visit_Lambda(self, node):
    return translate_function_ast("lambda", node.args, [ast.Return(node.body)], parent = self)
    
def translate_function_ast(name, args, body, 
                              globals_dict = None,
                              closure_vars = [], 
                              closure_cells = [],
                              parent = None, 
                              filename = None):
  """
  Helper to launch translation of a python function's AST, and then construct
  an untyped parakeet function from the arguments, refs, and translated body.
  """
  
  assert len(closure_vars) == len(closure_cells)
  closure_cell_dict = dict(zip(closure_vars, closure_cells))
  
  if filename is None and parent is not None:
    filename = parent.filename 
    
  translator = AST_Translator(globals_dict, closure_cell_dict, 
                              parent, function_name = name, filename = filename)

  assert not args.kwarg, "Parakeet doesn't support **kwargs, found in %s%s(%s)" % \
    (filename +":" if filename else "", name, args) 
  ssa_args, assignments = translator.translate_args(args)
  
  doc_string = None
  if len(body) > 0 and isinstance(body[0], ast.Expr):
    if isinstance(body[0].value, ast.Str):
      doc_string = body[0].value.s
      body = body[1:]
     
    
  _, body = translator.visit_block(body)
  body = assignments + body

  ssa_fn_name = names.fresh(name)

  # if function was nested in parakeet, it can have references to its
  # surrounding parakeet scope, which can't be captured with a python ref cell
  original_outer_names = translator.original_outer_names
  localized_outer_names = translator.localized_outer_names
  python_refs = translator.python_refs 
  ssa_args.prepend_nonlocal_args(localized_outer_names)
  if globals_dict:
    assert parent is None
    assert len(original_outer_names) == len(python_refs)
    return UntypedFn(ssa_fn_name, ssa_args, body, python_refs.values(), [])
  else:
    assert parent
    fn = UntypedFn(ssa_fn_name, ssa_args, body, [], original_outer_names, doc_string = doc_string)
    if len(original_outer_names) > 0:
      try: 
        outer_ssa_vars = [parent.lookup(x) for x in original_outer_names]
      except NameNotFound:
        print "Failure while trying to look up non-local names used by function %s" % name 
        raise 
      return syntax.Closure(fn, outer_ssa_vars)
    else:
      return fn

def strip_leading_whitespace(source):
  lines = source.splitlines()
  assert len(lines) > 0
  first_line = lines[0]
  n_removed = len(first_line) - len(first_line.lstrip())
  if n_removed > 0:
    return '\n'.join(line[n_removed:] for line in lines)
  else:
    return source

def translate_function_source(source, 
                              globals_dict, 
                              closure_vars = [],
                              closure_cells = [],
                              filename = None):
  assert len(closure_vars) == len(closure_cells)
  syntax_tree = ast.parse(strip_leading_whitespace(source))

  if isinstance(syntax_tree, (ast.Module, ast.Interactive)):
    assert len(syntax_tree.body) == 1
    syntax_tree = syntax_tree.body[0]
  elif isinstance(syntax_tree, ast.Expression):
    syntax_tree = syntax_tree.body
  
  assert isinstance(syntax_tree, ast.FunctionDef), \
      "Unexpected Python syntax node: %s" % ast.dump(syntax_tree)
  return translate_function_ast(syntax_tree.name, 
                                syntax_tree.args, 
                                syntax_tree.body, 
                                globals_dict, 
                                closure_vars,
                                closure_cells, 
                                filename = filename)

# python value of a user-defined function mapped to its
# untyped representation
_known_python_functions = {}

# keep track of which functions are being translated at this moment 
# to check for recursive calls 
_currently_processing = set([])


def _translate_function_value(fn):
  """
  The core of function translation, should only end up here 
  if the python function's intermediate representation isn't cached
  """
  assert fn not in _currently_processing, \
    "Recursion detected through function value %s" % (fn,)

  if isinstance(fn, Prim):
    fundef = build_untyped_prim_fn(fn)
  elif isinstance(fn, (Type, np.dtype, int, bool, long, float)):
    fundef = build_untyped_cast_fn(fn)
  elif isinstance(fn, type) and Expr in fn.mro():
    fundef = build_untyped_expr_fn(fn)  
  elif isinstance(fn, macro):
    fundef = fn.as_fn()
    
  else:
    # if it's not a macro or some sort of internal expression
    # then we're really dealing with a Python function
    # so get to work pulling apart its AST and translating
    # it into Parakeet IR
    assert type(fn) not in \
      (types.BuiltinFunctionType, types.TypeType, np.ufunc, types.MethodType), \
      "Unsupported function: %s" % (fn,) 
    
    _currently_processing.add(fn) 
    try:
      assert hasattr(fn, 'func_globals'), "Expected function to have globals: %s" % fn
      assert hasattr(fn, 'func_closure'), "Expected function to have closure cells: %s" % fn
      assert hasattr(fn, 'func_code'), "Expected function to have code object: %s" % fn

      source = inspect.getsource(fn)
      filename = inspect.getsourcefile(fn)
    except:
      _currently_processing.remove(fn)
      assert False, "Parakeet couldn't access source of function %s" % fn 
    globals_dict = fn.func_globals

    free_vars = fn.func_code.co_freevars
    closure_cells = fn.func_closure
    if closure_cells is None: 
      closure_cells = ()
    try: 
      fundef = translate_function_source(source,
                                        globals_dict,
                                        free_vars,
                                        closure_cells, 
                                        filename = filename)
    except:
      _currently_processing.remove(fn)
      raise 
    if config.print_untyped_function:
      print "[ast_conversion] Translated %s into untyped function:\n%s" % (fn, repr(fundef))
  
    _currently_processing.remove(fn)              
    _known_python_functions[fn] = fundef
  return fundef 

import threading 
_lock = threading.RLock()

def translate_function_value(fn):
  if fn in _known_python_functions:
    return _known_python_functions[fn]
  
  # if it's already a Parakeet function, just return it 
  if isinstance(fn, UntypedFn):
    return fn 
  
  # short-circuit logic for turning dtypes and Python types into 
  # functions for casting from any value to those types 
  elif isinstance(fn, (np.dtype, int, long, float, bool)): 
    fundef = build_untyped_cast_fn(fn)
      
  else:
    
    # none of the obvious shortcuts worked, 
    # we're going to have to translate this function for real 
    from ..mappings import function_mappings 
    if fn in function_mappings:
      fn = function_mappings[fn]
      
 
    # ...unless we forgot to add it to mappings but some equivalent primitive 
    # got registered 
    if fn in prims.prim_lookup_by_value:
      fn = prims.prim_lookup_by_value[fn]
  
    # if the function has been wrapped with a decorator, unwrap it 
    while isinstance(fn, jit):
      fn = fn.f
    
    # check again if, after unwrapping the function a bunch, it isn't in the cache 
    if fn in _known_python_functions:
      return _known_python_functions[fn]
  
    with _lock:
      fundef = _translate_function_value(fn)
           
  _known_python_functions[fn] = fundef 
  return fundef 
